"""from https://github.com/keithito/tacotron"""

"""
Cleaners are transformations that run over the input text at both training and eval time.

Cleaners can be selected by passing a comma-delimited list of cleaner names as the "cleaners"
hyperparameter. Some cleaners are English-specific. You'll typically want to use:
  1. "english_cleaners" for English text
  2. "transliteration_cleaners" for non-English text that can be transliterated to ASCII using
     the Unidecode library (https://pypi.python.org/pypi/Unidecode)
  3. "basic_cleaners" if you do not want to transliterate (in this case, you should also update
     the symbols in symbols.py to match your data).
"""


# Regular expression matching whitespace:


import re
import inflect
from unidecode import unidecode
import eng_to_ipa as ipa

_inflect = inflect.engine()
_comma_number_re = re.compile(r"([0-9][0-9\,]+[0-9])")
_decimal_number_re = re.compile(r"([0-9]+\.[0-9]+)")
_pounds_re = re.compile(r"£([0-9\,]*[0-9]+)")
_dollars_re = re.compile(r"\$([0-9]+(?:\.[0-9]+)?)")
_ordinal_re = re.compile(r"([0-9]+)(st|nd|rd|th)")
# Decade/century plurals like "1800s", "1990s", "60s", "'90s"
_decade_plural_re = re.compile(r"'?(\d{2,4})s\b")
# Captures number and suffix separately
_number_re = re.compile(r"[0-9]+")
# Time expressions like "9:16 AM" or "3:30 PM"
_time_re = re.compile(r"(\d{1,2}):(\d{2})\s*(AM|PM|am|pm|a\.m\.|p\.m\.)")
# Standalone AM/PM (after time processing or alone)
_am_pm_re = re.compile(r"\b(AM|PM|am|pm)\b")

# Custom word pronunciations for words the model mispronounces
# Format: word -> phonetic respelling that sounds correct
_custom_pronunciations = {
    "bowtie": "bow tie",
    "bowties": "bow ties",
    "I'll": "Aisle",
    "i'll": "aisle",
    "I'd": "Eyed",
    "i'd": "eyed",
}

# List of (regular expression, replacement) pairs for abbreviations:
_abbreviations = [
    (re.compile(r"\b" + re.escape(x[0]) + r"\.", re.IGNORECASE), x[1])
    for x in [
        ("mrs", "misess"),
        ("mr", "mister"),
        ("dr", "doctor"),
        ("st", "saint"),
        ("co", "company"),
        ("jr", "junior"),
        ("maj", "major"),
        ("gen", "general"),
        ("drs", "doctors"),
        ("rev", "reverend"),
        ("lt", "lieutenant"),
        ("hon", "honorable"),
        ("sgt", "sergeant"),
        ("capt", "captain"),
        ("esq", "esquire"),
        ("ltd", "limited"),
        ("col", "colonel"),
        ("ft", "fort"),
    ]
]


# List of (ipa, lazy ipa) pairs:
_lazy_ipa = [
    (re.compile("%s" % x[0]), x[1])
    for x in [
        ("r", "ɹ"),
        ("æ", "e"),
        ("ɑ", "a"),
        ("ɔ", "o"),
        ("ð", "z"),
        ("θ", "s"),
        ("ɛ", "e"),
        ("ɪ", "i"),
        ("ʊ", "u"),
        ("ʒ", "ʥ"),
        ("ʤ", "ʥ"),
        ("ˈ", "↓"),
    ]
]

# List of (ipa, lazy ipa2) pairs:
_lazy_ipa2 = [
    (re.compile("%s" % x[0]), x[1])
    for x in [
        ("r", "ɹ"),
        ("ð", "z"),
        ("θ", "s"),
        ("ʒ", "ʑ"),
        ("ʤ", "dʑ"),
        ("ˈ", "↓"),
    ]
]

# List of (ipa, ipa2) pairs
_ipa_to_ipa2 = [
    (re.compile("%s" % x[0]), x[1])
    for x in [("r", "ɹ"), ("ʤ", "dʒ"), ("ʧ", "tʃ")]
]


def apply_custom_pronunciations(text):
    """Replace words with custom pronunciations for better TTS output."""
    for word, replacement in _custom_pronunciations.items():
        # Case-insensitive replacement while preserving original case pattern
        pattern = re.compile(re.escape(word), re.IGNORECASE)
        text = pattern.sub(replacement, text)
    return text


def expand_abbreviations(text):
    for regex, replacement in _abbreviations:
        text = re.sub(regex, replacement, text)
    return text


def normalize_punctuation_for_pauses(text):
    """Convert dashes, ellipses, and colons to commas for natural pauses in speech."""
    # Em-dash (—) and en-dash (–) should create pauses
    text = re.sub(r"[—–]", ", ", text)
    # Ellipsis character (…) or three dots (...) should create longer pauses
    text = re.sub(r"…", "...", text)  # Normalize ellipsis character to dots
    text = re.sub(r"\.{3,}", ", , ", text)  # Multiple dots -> longer pause
    # Colon should create a pause (but not in time expressions like 9:16)
    text = re.sub(r"(?<!\d):(?!\d)", ", ", text)  # Colon not surrounded by digits
    # Clean up multiple commas/spaces
    text = re.sub(r",\s*,", ", ", text)
    text = re.sub(r"\s+", " ", text)
    return text


def collapse_whitespace(text):
    return re.sub(r"\s+", " ", text)


def _remove_commas(m):
    return m.group(1).replace(",", "")


def _expand_decimal_point(m):
    integer, decimal = m.group(1).split(".")
    return f"{integer} point {decimal}"


def _expand_dollars(m):
    match = m.group(1)
    parts = match.split(".")
    if len(parts) > 2:
        return match + " dollars"  # Unexpected format
    dollars = int(parts[0]) if parts[0] else 0
    cents = int(parts[1]) if len(parts) > 1 and parts[1] else 0
    if dollars and cents:
        dollar_unit = "dollar" if dollars == 1 else "dollars"
        cent_unit = "cent" if cents == 1 else "cents"
        return f"{dollars} {dollar_unit}, {cents} {cent_unit}"
    elif dollars:
        dollar_unit = "dollar" if dollars == 1 else "dollars"
        return f"{dollars} {dollar_unit}"
    elif cents:
        cent_unit = "cent" if cents == 1 else "cents"
        return f"{cents} {cent_unit}"
    else:
        return "zero dollars"


def _expand_pounds(m):
    num_str = m.group(1).replace(",", "")  # Remove commas from pound amount
    num = int(num_str)
    return f"{num} {'pound' if num == 1 else 'pounds'}"


def _expand_decade_plural(m):
    """Expand decade/century plurals like '1800s' -> 'eighteen hundreds'."""
    num_str = m.group(1)
    num = int(num_str)
    
    # Handle short decades like '60s, '90s or just 60s, 90s
    if num < 100:
        decade_word = _inflect.number_to_words(num, andword="")
        return f"{decade_word}s"  # "sixties", "nineties"
    
    # Handle century references like 1800s, 1900s (meaning the century/era)
    if num % 100 == 0 and 1000 <= num <= 2000:
        century = num // 100
        century_word = _inflect.number_to_words(century, andword="")
        return f"{century_word} hundreds"  # "eighteen hundreds"
    
    # Handle decade references like 1980s, 1990s, 2020s
    if num >= 1000:
        # Split into century and decade parts
        century_part = num // 100  # e.g., 19 for 1980
        decade_part = (num % 100) // 10 * 10  # e.g., 80 for 1980
        
        if century_part >= 20:  # 2000s and beyond
            if decade_part == 0:
                # 2000s -> "two thousands"
                return _inflect.number_to_words(num, andword="") + "s"
            else:
                # 2020s -> "twenty twenties"
                century_word = _inflect.number_to_words(century_part, andword="")
                decade_word = _inflect.number_to_words(decade_part, andword="")
                return f"{century_word} {decade_word}s"
        else:  # 1000s-1900s
            # 1980s -> "nineteen eighties"
            century_word = _inflect.number_to_words(century_part, andword="")
            decade_word = _inflect.number_to_words(decade_part, andword="")
            return f"{century_word} {decade_word}s"
    
    # Fallback for other numbers
    return _inflect.number_to_words(num, andword="") + "s"


def _expand_time(m):
    """Expand time expressions like '9:16 AM' -> 'nine sixteen A M'."""
    hour = int(m.group(1))
    minute = int(m.group(2))
    period = m.group(3).upper().replace(".", "")  # Normalize to AM or PM
    
    hour_word = _inflect.number_to_words(hour, andword="")
    
    if minute == 0:
        minute_word = "o'clock"
    elif minute < 10:
        minute_word = "oh " + _inflect.number_to_words(minute, andword="")
    else:
        minute_word = _inflect.number_to_words(minute, andword="")
    
    # Expand AM/PM to spelled out letters
    period_expanded = " ".join(period)  # "AM" -> "A M", "PM" -> "P M"
    
    return f"{hour_word} {minute_word} {period_expanded}"


def _expand_am_pm(m):
    """Expand standalone AM/PM to 'A M' / 'P M'."""
    period = m.group(1).upper()
    return " ".join(period)  # "AM" -> "A M", "PM" -> "P M"


def _expand_ordinal(m):
    num_str = m.group(1)  # This is the number part, e.g., "21" from "21st"
    # First get ordinal string (e.g., "21st"), then convert to words ("twenty-first")
    return _inflect.number_to_words(_inflect.ordinal(num_str))


def _expand_number(m):
    num = int(m.group(0))
    s_num = m.group(0)  # string form, useful for 'oh' in years like 1907

    # Default expansion using inflect
    expanded_num = _inflect.number_to_words(num, andword="").replace(",", "")

    # Specific year pronunciations overrides
    # Handles 19xx years
    if 1900 <= num <= 1999:
        if num % 100 == 0:  # e.g., 1900
            return (
                _inflect.number_to_words(num // 100, andword="") + " hundred"
            )  # "nineteen hundred"
        else:  # e.g., 1901, 1984
            part1 = _inflect.number_to_words(
                num // 100, andword=""
            )  # "nineteen"
            part2_val = num % 100
            if (
                s_num[2] == "0"
            ):  # Handles "oh" for years like 1907 ("nineteen oh seven")
                part2 = "oh " + _inflect.number_to_words(part2_val, andword="")
            else:  # Handles years like 1984 ("nineteen eighty-four")
                part2 = _inflect.number_to_words(part2_val, andword="")
            return f"{part1} {part2}"
    # Handles 2000
    elif num == 2000:
        return "two thousand"
    # Handles 200x years (e.g., 2001-2009)
    elif 2000 < num < 2010:
        return "two thousand " + _inflect.number_to_words(
            num % 100, andword=""
        )  # "two thousand one"

    # For other numbers (e.g., 1234) or years where standard expansion is preferred (e.g., 2024),
    # the default 'expanded_num' (e.g. "one thousand two hundred thirty-four", "two thousand twenty-four") is returned.
    return expanded_num


def normalize_numbers(text):
    placeholders = {}
    placeholder_idx = 0

    def idx_to_alpha(idx):
        # Convert 0 -> A, 1 -> B, ..., 25 -> Z, 26 -> AA, etc.
        s = ""
        while True:
            s = chr(ord("A") + (idx % 26)) + s
            idx = idx // 26 - 1
            if idx < 0:
                break
        return s

    def add_placeholder(value_func, match_obj):
        nonlocal placeholder_idx
        key = f"PHPLACEHOLDER_{idx_to_alpha(placeholder_idx)}"
        placeholders[key] = value_func(match_obj)
        placeholder_idx += 1
        return key

    # 1. Remove commas from multi-digit numbers (e.g., "1,234" -> "1234")
    text = re.sub(_comma_number_re, _remove_commas, text)

    # 1.3. Process time expressions FIRST (e.g., "9:16 AM" -> "nine sixteen A M")
    text = re.sub(
        _time_re, lambda m: add_placeholder(_expand_time, m), text
    )

    # 1.5. Process decade/century plurals BEFORE other number processing (e.g., "1800s" -> "eighteen hundreds")
    text = re.sub(
        _decade_plural_re, lambda m: add_placeholder(_expand_decade_plural, m), text
    )

    # 2. Process and placeholder pounds. Uses _expand_pounds.
    text = re.sub(
        _pounds_re, lambda m: add_placeholder(_expand_pounds, m), text
    )

    # 3. Process and placeholder dollars. Uses _expand_dollars.
    text = re.sub(
        _dollars_re, lambda m: add_placeholder(_expand_dollars, m), text
    )

    # 4. Process and placeholder decimals. Uses _expand_decimal_point.
    text = re.sub(
        _decimal_number_re,
        lambda m: add_placeholder(_expand_decimal_point, m),
        text,
    )

    # 5. Process and placeholder ordinals. Uses _expand_ordinal.
    text = re.sub(
        _ordinal_re, lambda m: add_placeholder(_expand_ordinal, m), text
    )

    # 6. Expand remaining numbers to words. Uses _expand_number.
    # These are numbers that weren't part of currency, decimals, or ordinals.
    text = re.sub(
        _number_re, lambda m: add_placeholder(_expand_number, m), text
    )

    # 6.5. Expand any remaining standalone AM/PM
    text = re.sub(
        _am_pm_re, lambda m: add_placeholder(_expand_am_pm, m), text
    )

    # 7. Restore placeholders in the correct order.
    for key in sorted(placeholders.keys(), key=len, reverse=True):
        text = text.replace(key, placeholders[key])

    return text


def mark_dark_l(text):
    return re.sub(
        r"l([^aeiouæɑɔəɛɪʊ ]*(?: |$))", lambda x: "ɫ" + x.group(1), text
    )


def english_to_ipa(text):
    text = apply_custom_pronunciations(text)
    text = normalize_punctuation_for_pauses(text)
    text = unidecode(text).lower()
    text = expand_abbreviations(text)
    text = normalize_numbers(text)
    phonemes = ipa.convert(text)
    phonemes = collapse_whitespace(phonemes)
    return phonemes


def english_to_lazy_ipa(text):
    text = english_to_ipa(text)
    for regex, replacement in _lazy_ipa:
        text = re.sub(regex, replacement, text)
    return text


def english_to_ipa2(text):
    text = english_to_ipa(text)
    text = mark_dark_l(text)
    for regex, replacement in _ipa_to_ipa2:
        text = re.sub(regex, replacement, text)
    return text.replace("...", "…")


def english_to_lazy_ipa2(text):
    text = english_to_ipa(text)
    for regex, replacement in _lazy_ipa2:
        text = re.sub(regex, replacement, text)
    return text



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'canvas.ui'
##
## Created by: Qt User Interface Compiler version 6.9.0
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QApplication, QFrame, QGridLayout, QHBoxLayout,
    QLabel, QPushButton, QSizePolicy, QSpacerItem,
    QSplitter, QVBoxLayout, QWidget)

from airunner.components.application.gui.widgets.slider.slider_widget import SliderWidget
from airunner.components.art.gui.widgets.canvas.custom_view import CustomGraphicsView
from airunner.components.art.gui.widgets.stablediffusion.stablediffusion_generator_form import StableDiffusionGeneratorForm
from airunner.components.art.gui.widgets.stablediffusion.stablediffusion_tool_tab_widget import StablediffusionToolTabWidget
import airunner.feather_rc

class Ui_canvas(object):
    def setupUi(self, canvas):
        if not canvas.objectName():
            canvas.setObjectName(u"canvas")
        canvas.resize(1027, 531)
        sizePolicy = QSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(canvas.sizePolicy().hasHeightForWidth())
        canvas.setSizePolicy(sizePolicy)
        canvas.setMinimumSize(QSize(0, 0))
        canvas.setStyleSheet(u"b")
        self.gridLayout = QGridLayout(canvas)
        self.gridLayout.setSpacing(0)
        self.gridLayout.setObjectName(u"gridLayout")
        self.gridLayout.setContentsMargins(0, 0, 0, 0)
        self.central_widget = QWidget(canvas)
        self.central_widget.setObjectName(u"central_widget")
        sizePolicy.setHeightForWidth(self.central_widget.sizePolicy().hasHeightForWidth())
        self.central_widget.setSizePolicy(sizePolicy)
        self.central_widget.setMinimumSize(QSize(0, 0))
        self.central_widget.setStyleSheet(u"")
        self.gridLayout_4 = QGridLayout(self.central_widget)
        self.gridLayout_4.setObjectName(u"gridLayout_4")
        self.gridLayout_4.setVerticalSpacing(0)
        self.gridLayout_4.setContentsMargins(0, 0, 0, 0)
        self.widget = QWidget(self.central_widget)
        self.widget.setObjectName(u"widget")
        sizePolicy1 = QSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Fixed)
        sizePolicy1.setHorizontalStretch(0)
        sizePolicy1.setVerticalStretch(0)
        sizePolicy1.setHeightForWidth(self.widget.sizePolicy().hasHeightForWidth())
        self.widget.setSizePolicy(sizePolicy1)
        self.widget.setMaximumSize(QSize(16777215, 50))
        self.verticalLayout = QVBoxLayout(self.widget)
        self.verticalLayout.setSpacing(0)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.verticalLayout.setContentsMargins(0, 0, 0, 0)
        self.widget1 = QWidget(self.widget)
        self.widget1.setObjectName(u"widget1")
        sizePolicy2 = QSizePolicy(QSizePolicy.Policy.MinimumExpanding, QSizePolicy.Policy.Fixed)
        sizePolicy2.setHorizontalStretch(0)
        sizePolicy2.setVerticalStretch(0)
        sizePolicy2.setHeightForWidth(self.widget1.sizePolicy().hasHeightForWidth())
        self.widget1.setSizePolicy(sizePolicy2)
        self.widget1.setMinimumSize(QSize(370, 50))
        self.widget1.setMaximumSize(QSize(16777215, 50))
        self.widget1.setBaseSize(QSize(0, 50))
        self.horizontalLayout_2 = QHBoxLayout(self.widget1)
        self.horizontalLayout_2.setSpacing(5)
        self.horizontalLayout_2.setObjectName(u"horizontalLayout_2")
        self.horizontalLayout_2.setContentsMargins(5, 0, 5, 1)
        self.prompt_editor_button = QPushButton(self.widget1)
        self.prompt_editor_button.setObjectName(u"prompt_editor_button")
        self.prompt_editor_button.setMinimumSize(QSize(30, 30))
        self.prompt_editor_button.setMaximumSize(QSize(30, 30))
        self.prompt_editor_button.setCursor(QCursor(Qt.CursorShape.PointingHandCursor))
        icon = QIcon()
        icon.addFile(u":/light/icons/feather/light/message-square.svg", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        self.prompt_editor_button.setIcon(icon)
        self.prompt_editor_button.setCheckable(True)

        self.horizontalLayout_2.addWidget(self.prompt_editor_button)

        self.art_tools_button = QPushButton(self.widget1)
        self.art_tools_button.setObjectName(u"art_tools_button")
        self.art_tools_button.setMinimumSize(QSize(30, 30))
        self.art_tools_button.setMaximumSize(QSize(30, 30))
        self.art_tools_button.setCursor(QCursor(Qt.CursorShape.PointingHandCursor))
        icon1 = QIcon()
        icon1.addFile(u":/light/icons/feather/light/tool.svg", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        self.art_tools_button.setIcon(icon1)
        self.art_tools_button.setCheckable(True)

        self.horizontalLayout_2.addWidget(self.art_tools_button)

        self.line_7 = QFrame(self.widget1)
        self.line_7.setObjectName(u"line_7")
        self.line_7.setFrameShape(QFrame.Shape.VLine)
        self.line_7.setFrameShadow(QFrame.Shadow.Sunken)

        self.horizontalLayout_2.addWidget(self.line_7)

        self.new_button = QPushButton(self.widget1)
        self.new_button.setObjectName(u"new_button")
        self.new_button.setMinimumSize(QSize(30, 30))
        self.new_button.setMaximumSize(QSize(30, 30))
        self.new_button.setCursor(QCursor(Qt.CursorShape.PointingHandCursor))
        icon2 = QIcon()
        icon2.addFile(u":/light/icons/feather/light/file-plus.svg", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        self.new_button.setIcon(icon2)
        self.new_button.setFlat(False)

        self.horizontalLayout_2.addWidget(self.new_button)

        self.open_art_document = QPushButton(self.widget1)
        self.open_art_document.setObjectName(u"open_art_document")
        self.open_art_document.setMinimumSize(QSize(30, 30))
        self.open_art_document.setMaximumSize(QSize(30, 30))
        self.open_art_document.setCursor(QCursor(Qt.CursorShape.PointingHandCursor))
        icon3 = QIcon()
        icon3.addFile(u":/light/icons/feather/light/folder.svg", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        self.open_art_document.setIcon(icon3)

        self.horizontalLayout_2.addWidget(self.open_art_document)

        self.save_art_document = QPushButton(self.widget1)
        self.save_art_document.setObjectName(u"save_art_document")
        self.save_art_document.setMinimumSize(QSize(30, 30))
        self.save_art_document.setMaximumSize(QSize(30, 30))
        self.save_art_document.setCursor(QCursor(Qt.CursorShape.PointingHandCursor))
        icon4 = QIcon()
        icon4.addFile(u":/light/icons/feather/light/save.svg", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        self.save_art_document.setIcon(icon4)

        self.horizontalLayout_2.addWidget(self.save_art_document)

        self.import_button = QPushButton(self.widget1)
        self.import_button.setObjectName(u"import_button")
        self.import_button.setMinimumSize(QSize(30, 30))
        self.import_button.setMaximumSize(QSize(30, 30))
        self.import_button.setCursor(QCursor(Qt.CursorShape.PointingHandCursor))
        icon5 = QIcon()
        icon5.addFile(u":/light/icons/feather/light/arrow-down.svg", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        self.import_button.setIcon(icon5)
        self.import_button.setFlat(False)

        self.horizontalLayout_2.addWidget(self.import_button)

        self.export_button = QPushButton(self.widget1)
        self.export_button.setObjectName(u"export_button")
        self.export_button.setMinimumSize(QSize(30, 30))
        self.export_button.setMaximumSize(QSize(30, 30))
        self.export_button.setCursor(QCursor(Qt.CursorShape.PointingHandCursor))
        icon6 = QIcon()
        icon6.addFile(u":/light/icons/feather/light/arrow-up.svg", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        self.export_button.setIcon(icon6)
        self.export_button.setFlat(False)

        self.horizontalLayout_2.addWidget(self.export_button)

        self.line = QFrame(self.widget1)
        self.line.setObjectName(u"line")
        self.line.setFrameShape(QFrame.Shape.VLine)
        self.line.setFrameShadow(QFrame.Shadow.Sunken)

        self.horizontalLayout_2.addWidget(self.line)

        self.recenter_button = QPushButton(self.widget1)
        self.recenter_button.setObjectName(u"recenter_button")
        self.recenter_button.setMinimumSize(QSize(30, 30))
        self.recenter_button.setMaximumSize(QSize(30, 30))
        self.recenter_button.setCursor(QCursor(Qt.CursorShape.PointingHandCursor))
        icon7 = QIcon()
        icon7.addFile(u":/light/icons/feather/light/target.svg", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        self.recenter_button.setIcon(icon7)
        self.recenter_button.setFlat(False)

        self.horizontalLayout_2.addWidget(self.recenter_button)

        self.active_grid_area_button = QPushButton(self.widget1)
        self.active_grid_area_button.setObjectName(u"active_grid_area_button")
        self.active_grid_area_button.setMinimumSize(QSize(30, 30))
        self.active_grid_area_button.setMaximumSize(QSize(30, 30))
        self.active_grid_area_button.setCursor(QCursor(Qt.CursorShape.PointingHandCursor))
        icon8 = QIcon()
        icon8.addFile(u":/light/icons/feather/light/object-selected-icon.svg", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        self.active_grid_area_button.setIcon(icon8)
        self.active_grid_area_button.setCheckable(True)
        self.active_grid_area_button.setFlat(False)

        self.horizontalLayout_2.addWidget(self.active_grid_area_button)

        self.move_button = QPushButton(self.widget1)
        self.move_button.setObjectName(u"move_button")
        self.move_button.setMinimumSize(QSize(30, 30))
        self.move_button.setMaximumSize(QSize(30, 30))
        self.move_button.setCursor(QCursor(Qt.CursorShape.PointingHandCursor))
        icon9 = QIcon()
        icon9.addFile(u":/light/icons/feather/light/move.svg", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        self.move_button.setIcon(icon9)
        self.move_button.setCheckable(True)

        self.horizontalLayout_2.addWidget(self.move_button)

        self.line_4 = QFrame(self.widget1)
        self.line_4.setObjectName(u"line_4")
        self.line_4.setFrameShape(QFrame.Shape.VLine)
        self.line_4.setFrameShadow(QFrame.Shadow.Sunken)

        self.horizontalLayout_2.addWidget(self.line_4)

        self.filter_button = QPushButton(self.widget1)
        self.filter_button.setObjectName(u"filter_button")
        self.filter_button.setMinimumSize(QSize(30, 30))
        self.filter_button.setMaximumSize(QSize(30, 30))
        self.filter_button.setCursor(QCursor(Qt.CursorShape.PointingHandCursor))
        icon10 = QIcon()
        icon10.addFile(u":/light/icons/feather/light/filter.svg", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        self.filter_button.setIcon(icon10)

        self.horizontalLayout_2.addWidget(self.filter_button)

        self.brush_button = QPushButton(self.widget1)
        self.brush_button.setObjectName(u"brush_button")
        self.brush_button.setMinimumSize(QSize(30, 30))
        self.brush_button.setMaximumSize(QSize(30, 30))
        self.brush_button.setCursor(QCursor(Qt.CursorShape.PointingHandCursor))
        icon11 = QIcon()
        icon11.addFile(u":/light/icons/feather/light/pencil-icon.svg", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        self.brush_button.setIcon(icon11)
        self.brush_button.setCheckable(True)
        self.brush_button.setFlat(False)

        self.horizontalLayout_2.addWidget(self.brush_button)

        self.eraser_button = QPushButton(self.widget1)
        self.eraser_button.setObjectName(u"eraser_button")
        self.eraser_button.setMinimumSize(QSize(30, 30))
        self.eraser_button.setMaximumSize(QSize(30, 30))
        self.eraser_button.setCursor(QCursor(Qt.CursorShape.PointingHandCursor))
        icon12 = QIcon()
        icon12.addFile(u":/light/icons/feather/light/eraser-icon.svg", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        self.eraser_button.setIcon(icon12)
        self.eraser_button.setCheckable(True)
        self.eraser_button.setFlat(False)

        self.horizontalLayout_2.addWidget(self.eraser_button)

        self.line_5 = QFrame(self.widget1)
        self.line_5.setObjectName(u"line_5")
        self.line_5.setFrameShape(QFrame.Shape.VLine)
        self.line_5.setFrameShadow(QFrame.Shadow.Sunken)

        self.horizontalLayout_2.addWidget(self.line_5)

        self.grid_button = QPushButton(self.widget1)
        self.grid_button.setObjectName(u"grid_button")
        self.grid_button.setMinimumSize(QSize(30, 30))
        self.grid_button.setMaximumSize(QSize(30, 30))
        self.grid_button.setCursor(QCursor(Qt.CursorShape.PointingHandCursor))
        icon13 = QIcon()
        icon13.addFile(u":/light/icons/feather/light/grid.svg", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        self.grid_button.setIcon(icon13)
        self.grid_button.setCheckable(True)
        self.grid_button.setFlat(False)

        self.horizontalLayout_2.addWidget(self.grid_button)

        self.snap_to_grid_button = QPushButton(self.widget1)
        self.snap_to_grid_button.setObjectName(u"snap_to_grid_button")
        self.snap_to_grid_button.setMinimumSize(QSize(30, 30))
        self.snap_to_grid_button.setMaximumSize(QSize(30, 30))
        self.snap_to_grid_button.setCursor(QCursor(Qt.CursorShape.PointingHandCursor))
        icon14 = QIcon()
        icon14.addFile(u":/light/icons/feather/light/link-2.svg", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        self.snap_to_grid_button.setIcon(icon14)
        self.snap_to_grid_button.setCheckable(True)

        self.horizontalLayout_2.addWidget(self.snap_to_grid_button)

        self.line_2 = QFrame(self.widget1)
        self.line_2.setObjectName(u"line_2")
        self.line_2.setFrameShape(QFrame.Shape.VLine)
        self.line_2.setFrameShadow(QFrame.Shadow.Sunken)

        self.horizontalLayout_2.addWidget(self.line_2)

        self.undo_button = QPushButton(self.widget1)
        self.undo_button.setObjectName(u"undo_button")
        self.undo_button.setMinimumSize(QSize(30, 30))
        self.undo_button.setMaximumSize(QSize(30, 30))
        self.undo_button.setCursor(QCursor(Qt.CursorShape.PointingHandCursor))
        icon15 = QIcon()
        icon15.addFile(u":/light/icons/feather/light/corner-up-left.svg", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        self.undo_button.setIcon(icon15)
        self.undo_button.setFlat(False)

        self.horizontalLayout_2.addWidget(self.undo_button)

        self.redo_button = QPushButton(self.widget1)
        self.redo_button.setObjectName(u"redo_button")
        self.redo_button.setMinimumSize(QSize(30, 30))
        self.redo_button.setMaximumSize(QSize(30, 30))
        self.redo_button.setCursor(QCursor(Qt.CursorShape.PointingHandCursor))
        icon16 = QIcon()
        icon16.addFile(u":/light/icons/feather/light/corner-up-right.svg", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        self.redo_button.setIcon(icon16)
        self.redo_button.setFlat(False)

        self.horizontalLayout_2.addWidget(self.redo_button)

        self.horizontalSpacer = QSpacerItem(425, 9, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.horizontalLayout_2.addItem(self.horizontalSpacer)

        self.brush_size_slider = SliderWidget(self.widget1)
        self.brush_size_slider.setObjectName(u"brush_size_slider")
        self.brush_size_slider.setMinimumSize(QSize(100, 40))
        self.brush_size_slider.setMaximumSize(QSize(16777215, 40))
        self.brush_size_slider.setProperty(u"slider_minimum", 1)
        self.brush_size_slider.setProperty(u"slider_maximum", 100)
        self.brush_size_slider.setProperty(u"spinbox_minimum", 1)
        self.brush_size_slider.setProperty(u"spinbox_maximum", 100)
        self.brush_size_slider.setProperty(u"slider_tick_interval", 1)
        self.brush_size_slider.setProperty(u"slider_single_step", 1)
        self.brush_size_slider.setProperty(u"slider_page_step", 10)
        self.brush_size_slider.setProperty(u"spinbox_single_step", 1)
        self.brush_size_slider.setProperty(u"spinbox_page_step", 10)
        self.brush_size_slider.setProperty(u"settings_property", u"brush_settings.size")
        self.brush_size_slider.setProperty(u"display_as_float", False)

        self.horizontalLayout_2.addWidget(self.brush_size_slider)

        self.brush_color_button = QPushButton(self.widget1)
        self.brush_color_button.setObjectName(u"brush_color_button")
        self.brush_color_button.setMinimumSize(QSize(30, 30))
        self.brush_color_button.setMaximumSize(QSize(30, 30))
        self.brush_color_button.setCursor(QCursor(Qt.CursorShape.PointingHandCursor))

        self.horizontalLayout_2.addWidget(self.brush_color_button)


        self.verticalLayout.addWidget(self.widget1)

        self.line_3 = QFrame(self.widget)
        self.line_3.setObjectName(u"line_3")
        self.line_3.setFrameShape(QFrame.Shape.HLine)
        self.line_3.setFrameShadow(QFrame.Shadow.Sunken)

        self.verticalLayout.addWidget(self.line_3)


        self.gridLayout_4.addWidget(self.widget, 0, 0, 1, 1)

        self.splitter = QSplitter(self.central_widget)
        self.splitter.setObjectName(u"splitter")
        self.splitter.setOrientation(Qt.Orientation.Horizontal)
        self.prompts = StableDiffusionGeneratorForm(self.splitter)
        self.prompts.setObjectName(u"prompts")
        sizePolicy3 = QSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Preferred)
        sizePolicy3.setHorizontalStretch(0)
        sizePolicy3.setVerticalStretch(0)
        sizePolicy3.setHeightForWidth(self.prompts.sizePolicy().hasHeightForWidth())
        self.prompts.setSizePolicy(sizePolicy3)
        self.prompts.setMinimumSize(QSize(350, 0))
        self.prompts.setMaximumSize(QSize(350, 16777215))
        self.prompts.setBaseSize(QSize(350, 0))
        self.prompts.setStyleSheet(u"")
        self.splitter.addWidget(self.prompts)
        self.layoutWidget = QWidget(self.splitter)
        self.layoutWidget.setObjectName(u"layoutWidget")
        self.gridLayout_3 = QGridLayout(self.layoutWidget)
        self.gridLayout_3.setObjectName(u"gridLayout_3")
        self.gridLayout_3.setVerticalSpacing(0)
        self.gridLayout_3.setContentsMargins(0, 0, 0, 0)
        self.canvas_container = CustomGraphicsView(self.layoutWidget)
        self.canvas_container.setObjectName(u"canvas_container")
        sizePolicy4 = QSizePolicy(QSizePolicy.Policy.MinimumExpanding, QSizePolicy.Policy.Expanding)
        sizePolicy4.setHorizontalStretch(0)
        sizePolicy4.setVerticalStretch(0)
        sizePolicy4.setHeightForWidth(self.canvas_container.sizePolicy().hasHeightForWidth())
        self.canvas_container.setSizePolicy(sizePolicy4)
        self.canvas_container.setMinimumSize(QSize(1, 0))
        self.canvas_container.setBaseSize(QSize(1024, 0))
        self.canvas_container.setVerticalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        self.canvas_container.setHorizontalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)

        self.gridLayout_3.addWidget(self.canvas_container, 0, 0, 1, 1)

        self.horizontalLayout_4 = QHBoxLayout()
        self.horizontalLayout_4.setObjectName(u"horizontalLayout_4")
        self.horizontalLayout_4.setContentsMargins(10, 10, 10, 10)
        self.grid_info = QLabel(self.layoutWidget)
        self.grid_info.setObjectName(u"grid_info")
        sizePolicy1.setHeightForWidth(self.grid_info.sizePolicy().hasHeightForWidth())
        self.grid_info.setSizePolicy(sizePolicy1)
        font = QFont()
        font.setPointSize(10)
        self.grid_info.setFont(font)

        self.horizontalLayout_4.addWidget(self.grid_info)


        self.gridLayout_3.addLayout(self.horizontalLayout_4, 2, 0, 1, 1)

        self.line_6 = QFrame(self.layoutWidget)
        self.line_6.setObjectName(u"line_6")
        self.line_6.setFrameShape(QFrame.Shape.HLine)
        self.line_6.setFrameShadow(QFrame.Shadow.Sunken)

        self.gridLayout_3.addWidget(self.line_6, 1, 0, 1, 1)

        self.splitter.addWidget(self.layoutWidget)
        self.widget2 = StablediffusionToolTabWidget(self.splitter)
        self.widget2.setObjectName(u"widget2")
        sizePolicy3.setHeightForWidth(self.widget2.sizePolicy().hasHeightForWidth())
        self.widget2.setSizePolicy(sizePolicy3)
        self.widget2.setMinimumSize(QSize(0, 0))
        self.widget2.setMaximumSize(QSize(1024, 16777215))
        self.widget2.setBaseSize(QSize(0, 0))
        self.splitter.addWidget(self.widget2)

        self.gridLayout_4.addWidget(self.splitter, 1, 0, 1, 1)


        self.gridLayout.addWidget(self.central_widget, 0, 0, 1, 1)


        self.retranslateUi(canvas)

        QMetaObject.connectSlotsByName(canvas)
    # setupUi

    def retranslateUi(self, canvas):
        canvas.setWindowTitle(QCoreApplication.translate("canvas", u"Form", None))
        self.prompt_editor_button.setText("")
        self.art_tools_button.setText("")
#if QT_CONFIG(tooltip)
        self.new_button.setToolTip(QCoreApplication.translate("canvas", u"New art document", None))
#endif // QT_CONFIG(tooltip)
        self.new_button.setText("")
#if QT_CONFIG(tooltip)
        self.open_art_document.setToolTip(QCoreApplication.translate("canvas", u"Open art document", None))
#endif // QT_CONFIG(tooltip)
        self.open_art_document.setText("")
#if QT_CONFIG(tooltip)
        self.save_art_document.setToolTip(QCoreApplication.translate("canvas", u"Save art document", None))
#endif // QT_CONFIG(tooltip)
        self.save_art_document.setText("")
#if QT_CONFIG(tooltip)
        self.import_button.setToolTip(QCoreApplication.translate("canvas", u"Import image", None))
#endif // QT_CONFIG(tooltip)
        self.import_button.setText("")
#if QT_CONFIG(tooltip)
        self.export_button.setToolTip(QCoreApplication.translate("canvas", u"Export image", None))
#endif // QT_CONFIG(tooltip)
        self.export_button.setText("")
#if QT_CONFIG(tooltip)
        self.recenter_button.setToolTip(QCoreApplication.translate("canvas", u"Recenter grid", None))
#endif // QT_CONFIG(tooltip)
        self.recenter_button.setText("")
#if QT_CONFIG(tooltip)
        self.active_grid_area_button.setToolTip(QCoreApplication.translate("canvas", u"Move active grid area", None))
#endif // QT_CONFIG(tooltip)
        self.active_grid_area_button.setText("")
#if QT_CONFIG(tooltip)
        self.move_button.setToolTip(QCoreApplication.translate("canvas", u"Move layer tool", None))
#endif // QT_CONFIG(tooltip)
        self.move_button.setText("")
#if QT_CONFIG(tooltip)
        self.filter_button.setToolTip(QCoreApplication.translate("canvas", u"Toggle filters", None))
#endif // QT_CONFIG(tooltip)
        self.filter_button.setText("")
#if QT_CONFIG(tooltip)
        self.brush_button.setToolTip(QCoreApplication.translate("canvas", u"Toggle brush tool", None))
#endif // QT_CONFIG(tooltip)
        self.brush_button.setText("")
#if QT_CONFIG(tooltip)
        self.eraser_button.setToolTip(QCoreApplication.translate("canvas", u"Toggle earser tool", None))
#endif // QT_CONFIG(tooltip)
        self.eraser_button.setText("")
#if QT_CONFIG(tooltip)
        self.grid_button.setToolTip(QCoreApplication.translate("canvas", u"Toggle grid", None))
#endif // QT_CONFIG(tooltip)
        self.grid_button.setText("")
#if QT_CONFIG(tooltip)
        self.snap_to_grid_button.setToolTip(QCoreApplication.translate("canvas", u"Snap to grid", None))
#endif // QT_CONFIG(tooltip)
        self.snap_to_grid_button.setText("")
#if QT_CONFIG(tooltip)
        self.undo_button.setToolTip(QCoreApplication.translate("canvas", u"Undo", None))
#endif // QT_CONFIG(tooltip)
        self.undo_button.setText("")
#if QT_CONFIG(tooltip)
        self.redo_button.setToolTip(QCoreApplication.translate("canvas", u"Redo", None))
#endif // QT_CONFIG(tooltip)
        self.redo_button.setText("")
#if QT_CONFIG(tooltip)
        self.brush_size_slider.setToolTip(QCoreApplication.translate("canvas", u"Brush Size", None))
#endif // QT_CONFIG(tooltip)
        self.brush_size_slider.setProperty(u"label_text", "")
#if QT_CONFIG(tooltip)
        self.brush_color_button.setToolTip(QCoreApplication.translate("canvas", u"Brush Color", None))
#endif // QT_CONFIG(tooltip)
        self.brush_color_button.setText("")
        self.canvas_container.setProperty(u"canvas_type", QCoreApplication.translate("canvas", u"brush", None))
        self.grid_info.setText(QCoreApplication.translate("canvas", u"TextLabel", None))
    # retranslateUi




"""Unit tests for MultiFileCodeTool and MultiFileCodeSession."""

import pytest

from airunner.components.llm.tools.multi_file_code_tool import (
    MultiFileCodeSession,
    MultiFileCodeTool,
    FileSpec,
)
from airunner.components.llm.tools.code_session import (
    CodeSessionConfig,
)
from airunner.components.document_editor.workspace_manager import (
    WorkspaceManager,
)
from airunner.enums import CodeOperationType


@pytest.fixture
def temp_workspace(tmp_path):
    """Create temporary workspace directory."""
    workspace_dir = tmp_path / "test_workspace"
    workspace_dir.mkdir()
    return workspace_dir


@pytest.fixture
def workspace_manager(temp_workspace):
    """Create WorkspaceManager instance."""
    return WorkspaceManager(str(temp_workspace))


@pytest.fixture
def multi_file_tool(temp_workspace):
    """Create MultiFileCodeTool instance."""
    return MultiFileCodeTool(str(temp_workspace))


def test_file_spec_creation():
    """Test FileSpec dataclass."""
    spec = FileSpec(
        rel_path="test.py",
        operation=CodeOperationType.CREATE,
        initial_content="# Test",
    )

    assert spec.rel_path == "test.py"
    assert spec.operation == CodeOperationType.CREATE
    assert spec.initial_content == "# Test"


def test_multi_file_session_creation(workspace_manager):
    """Test MultiFileCodeSession initialization."""
    file_specs = [
        FileSpec("file1.py"),
        FileSpec("file2.py"),
        FileSpec("file3.py"),
    ]

    session = MultiFileCodeSession(
        multi_session_id="test_001",
        workspace=workspace_manager,
        file_specs=file_specs,
    )

    assert session.multi_session_id == "test_001"
    assert len(session.sessions) == 3
    assert "file1.py" in session.sessions
    assert "file2.py" in session.sessions
    assert "file3.py" in session.sessions


def test_receive_token_to_specific_file(workspace_manager):
    """Test sending tokens to a specific file."""
    file_specs = [
        FileSpec("file1.py"),
        FileSpec("file2.py"),
    ]

    session = MultiFileCodeSession(
        multi_session_id="test_002",
        workspace=workspace_manager,
        file_specs=file_specs,
    )

    # Send tokens to file1
    session.receive_token("file1.py", "import")
    session.receive_token("file1.py", " sys\n")

    # Send tokens to file2
    session.receive_token("file2.py", "import")
    session.receive_token("file2.py", " os\n")

    # Check accumulated content
    file1_session = session.get_session("file1.py")
    file2_session = session.get_session("file2.py")

    assert "import sys" in file1_session.get_accumulated_code()
    assert "import os" in file2_session.get_accumulated_code()


def test_complete_single_file(workspace_manager):
    """Test completing a single file in multi-file session."""
    file_specs = [
        FileSpec("file1.py"),
        FileSpec("file2.py"),
    ]

    config = CodeSessionConfig(
        flush_token_count=100,  # High threshold to avoid auto-flush
        auto_format=False,
    )

    session = MultiFileCodeSession(
        multi_session_id="test_003",
        workspace=workspace_manager,
        file_specs=file_specs,
        config=config,
    )

    # Add content to file1
    session.receive_token("file1.py", "print('hello')")

    # Complete file1
    abs_path = session.complete_file("file1.py")

    assert abs_path is not None
    assert abs_path.endswith("file1.py")
    assert session.is_complete("file1.py")
    assert not session.is_complete("file2.py")
    assert not session.are_all_complete()

    # Verify file exists
    assert workspace_manager.exists("file1.py")


def test_complete_all_files(workspace_manager):
    """Test completing all files in multi-file session."""
    file_specs = [
        FileSpec("file1.py"),
        FileSpec("file2.py"),
        FileSpec("file3.py"),
    ]

    config = CodeSessionConfig(flush_token_count=100, auto_format=False)

    session = MultiFileCodeSession(
        multi_session_id="test_004",
        workspace=workspace_manager,
        file_specs=file_specs,
        config=config,
    )

    # Add content to all files
    session.receive_token("file1.py", "# File 1")
    session.receive_token("file2.py", "# File 2")
    session.receive_token("file3.py", "# File 3")

    # Complete all
    completed = session.complete_all()

    assert len(completed) == 3
    assert session.are_all_complete()

    # Verify all files exist
    assert workspace_manager.exists("file1.py")
    assert workspace_manager.exists("file2.py")
    assert workspace_manager.exists("file3.py")


def test_file_flush_callback(workspace_manager):
    """Test file flush callback is called with correct data."""
    flush_events = []

    def on_file_flush(file_path: str, session_id: str, content: str):
        flush_events.append((file_path, session_id, content))

    file_specs = [FileSpec("file1.py")]

    config = CodeSessionConfig(flush_token_count=5)

    session = MultiFileCodeSession(
        multi_session_id="test_005",
        workspace=workspace_manager,
        file_specs=file_specs,
        config=config,
        on_file_flush=on_file_flush,
    )

    # Send enough tokens to trigger flush (5 chars)
    for char in "import sys\n":
        session.receive_token("file1.py", char)

    # Should have triggered at least one flush
    assert len(flush_events) >= 1
    assert flush_events[0][0] == "file1.py"
    # After 5 tokens we get "impor"
    assert "impor" in flush_events[0][2]


def test_file_complete_callback(workspace_manager):
    """Test file completion callback."""
    complete_events = []

    def on_file_complete(file_path: str, session_id: str):
        complete_events.append((file_path, session_id))

    file_specs = [FileSpec("file1.py")]

    config = CodeSessionConfig(auto_format=False)

    session = MultiFileCodeSession(
        multi_session_id="test_006",
        workspace=workspace_manager,
        file_specs=file_specs,
        config=config,
        on_file_complete=on_file_complete,
    )

    session.receive_token("file1.py", "code")
    session.complete_file("file1.py")

    assert len(complete_events) == 1
    assert complete_events[0][0] == "file1.py"


def test_multi_file_tool_create_session(multi_file_tool):
    """Test MultiFileCodeTool session creation."""
    file_specs = [
        FileSpec("app.py"),
        FileSpec("utils.py"),
    ]

    session = multi_file_tool.create_multi_file_session(file_specs)

    assert session is not None
    assert session.multi_session_id.startswith("multi_code_")
    assert len(session.get_file_paths()) == 2


def test_multi_file_tool_get_session(multi_file_tool):
    """Test retrieving sessions from tool."""
    file_specs = [FileSpec("test.py")]

    session = multi_file_tool.create_multi_file_session(
        file_specs, multi_session_id="custom_id"
    )

    retrieved = multi_file_tool.get_multi_session("custom_id")
    assert retrieved is session


def test_multi_file_tool_remove_session(multi_file_tool):
    """Test removing completed sessions."""
    file_specs = [FileSpec("test.py")]

    session = multi_file_tool.create_multi_file_session(
        file_specs, multi_session_id="remove_me"
    )

    # Remove it
    multi_file_tool.remove_multi_session("remove_me")

    # Should be gone
    assert multi_file_tool.get_multi_session("remove_me") is None


def test_multi_file_tool_active_sessions(multi_file_tool):
    """Test filtering active sessions."""
    file_specs1 = [FileSpec("file1.py")]
    file_specs2 = [FileSpec("file2.py")]

    config = CodeSessionConfig(auto_format=False)

    session1 = multi_file_tool.create_multi_file_session(
        file_specs1,
        multi_session_id="active1",
        config=config,
    )

    session2 = multi_file_tool.create_multi_file_session(
        file_specs2,
        multi_session_id="active2",
        config=config,
    )

    # Both active initially
    active = multi_file_tool.get_active_multi_sessions()
    assert len(active) == 2

    # Complete session1
    session1.receive_token("file1.py", "code")
    session1.complete_all()

    # Only session2 should be active
    active = multi_file_tool.get_active_multi_sessions()
    assert len(active) == 1
    assert active[0].multi_session_id == "active2"


def test_initial_content(workspace_manager):
    """Test file specs with initial content."""
    file_specs = [
        FileSpec("file1.py", initial_content="# Header\n"),
    ]

    config = CodeSessionConfig(flush_token_count=100, auto_format=False)

    session = MultiFileCodeSession(
        multi_session_id="test_007",
        workspace=workspace_manager,
        file_specs=file_specs,
        config=config,
    )

    # Add more content
    session.receive_token("file1.py", "print('hello')")

    # Complete
    session.complete_file("file1.py")

    # Check content includes initial and added content
    content = workspace_manager.read_file("file1.py")
    assert "# Header" in content
    assert "print('hello')" in content


def test_invalid_file_path(workspace_manager):
    """Test handling of invalid file path."""
    file_specs = [FileSpec("file1.py")]

    session = MultiFileCodeSession(
        multi_session_id="test_008",
        workspace=workspace_manager,
        file_specs=file_specs,
    )

    # Try to send token to non-existent file
    session.receive_token("nonexistent.py", "code")

    # Should not crash, just log warning
    assert session.get_session("nonexistent.py") is None



"""Pytest configuration for widgets tests."""

import pytest
from PySide6.QtWidgets import QApplication


@pytest.fixture(scope="session")
def qapp():
    """Create QApplication instance for the test session."""
    app = QApplication.instance()
    if app is None:
        app = QApplication([])
    yield app
    # Don't quit - let pytest handle cleanup



"""Mixin for managing canvas items and their positioning."""

from typing import Optional

from PySide6.QtGui import QImage

from airunner.components.art.gui.widgets.canvas.draggables.layer_image_item import (
    LayerImageItem,
)


class CanvasItemManagementMixin:
    """Handles creation and management of canvas items.

    This mixin provides methods for creating, updating, and managing
    LayerImageItem instances on the canvas.
    """

    def _create_new_item(self, image: QImage, x: int, y: int) -> None:
        """Create a new LayerImageItem and add it to the scene.

        Args:
            image: The QImage to display.
            x: X position for the item.
            y: Y position for the item.
        """
        self.item = LayerImageItem(image)
        if self.item.scene() is None:
            self.addItem(self.item)
            self.item.setPos(x, y)
            self.original_item_positions[self.item] = self.item.pos()

    def _update_existing_item(self, image: QImage, x: int, y: int) -> None:
        """Update an existing item with new image and position.

        Args:
            image: The QImage to display.
            x: X position for the item.
            y: Y position for the item.
        """
        self.item.setPos(x, y)
        self.original_item_positions[self.item] = self.item.pos()
        if image is not None and not image.isNull():
            try:
                self.item.updateImage(image)
            except Exception:
                self.logger.warning(
                    "Failed to update existing item with new image."
                )

    def set_item(
        self,
        image: Optional[QImage] = None,
        z_index: int = 5,
        x: Optional[int] = None,
        y: Optional[int] = None,
    ) -> None:
        """Create or update the main canvas item with an image.

        Args:
            image: The QImage to display. If None, no item is created.
            z_index: The z-order for the item (default: 5).
            x: X position for the item. If None, uses grid settings.
            y: Y position for the item. If None, uses grid settings.
        """
        self.setSceneRect(self._extended_viewport_rect)

        if image is not None:
            x = self.active_grid_settings.pos_x if x is None else x
            y = self.active_grid_settings.pos_y if y is None else y

            # Check if we have layer items - if so, don't use the old drawing pad item system
            if len(self._layer_items) > 0:
                return

            if self.item is None:
                self._create_new_item(image, x, y)
            else:
                self._update_existing_item(image, x, y)

            self.item.setZValue(z_index)
            self.item.setVisible(True)

    def clear_selection(self) -> None:
        """Clear both selection start and stop positions."""
        self.selection_start_pos = None
        self.selection_stop_pos = None



"""Evaluation testing framework components."""

from airunner.components.eval.client import AIRunnerClient
from airunner.components.eval.evaluators import (
    LLMAsJudge,
    create_correctness_evaluator,
    create_conciseness_evaluator,
    create_helpfulness_evaluator,
    create_relevance_evaluator,
)
from airunner.components.eval.datasets import (
    MATH_REASONING_DATASET,
    GENERAL_KNOWLEDGE_DATASET,
    SUMMARIZATION_DATASET,
    CODING_DATASET,
    INSTRUCTION_FOLLOWING_DATASET,
    REASONING_DATASET,
    get_dataset_by_category,
    get_dataset_by_difficulty,
    get_all_test_cases,
)

__all__ = [
    "AIRunnerClient",
    "LLMAsJudge",
    "create_correctness_evaluator",
    "create_conciseness_evaluator",
    "create_helpfulness_evaluator",
    "create_relevance_evaluator",
    "MATH_REASONING_DATASET",
    "GENERAL_KNOWLEDGE_DATASET",
    "SUMMARIZATION_DATASET",
    "CODING_DATASET",
    "INSTRUCTION_FOLLOWING_DATASET",
    "REASONING_DATASET",
    "get_dataset_by_category",
    "get_dataset_by_difficulty",
    "get_all_test_cases",
]



"""LangChain tools for calendar event management."""

from datetime import datetime
from typing import Optional
from langchain_core.tools import BaseTool
from pydantic import BaseModel as PydanticBaseModel, Field
from airunner.components.calendar.data.event import Event
from airunner.components.calendar.data.reminder import Reminder
from airunner.components.calendar.data.recurring_event import RecurringEvent
from airunner.components.data.session_manager import session_scope


class CreateEventInput(PydanticBaseModel):
    """Input schema for create_event tool."""

    title: str = Field(description="Event title/name")
    description: Optional[str] = Field(
        default=None, description="Detailed event description"
    )
    start_time: str = Field(
        description="Start time in ISO format (YYYY-MM-DDTHH:MM:SS)"
    )
    end_time: str = Field(
        description="End time in ISO format (YYYY-MM-DDTHH:MM:SS)"
    )
    location: Optional[str] = Field(default=None, description="Event location")
    category: Optional[str] = Field(
        default=None,
        description="Event category (e.g., work, personal, meeting)",
    )
    all_day: bool = Field(
        default=False, description="Whether event is all-day"
    )


class CreateEventTool(BaseTool):
    """Tool for creating calendar events."""

    name: str = "create_event"
    description: str = (
        "Create a new calendar event with specified date, time, "
        "title, and optional description. Use ISO format for times "
        "(YYYY-MM-DDTHH:MM:SS)."
    )
    args_schema: type[PydanticBaseModel] = CreateEventInput

    def _run(
        self,
        title: str,
        start_time: str,
        end_time: str,
        description: Optional[str] = None,
        location: Optional[str] = None,
        category: Optional[str] = None,
        all_day: bool = False,
    ) -> str:
        """Create a calendar event.

        Args:
            title: Event title
            start_time: ISO format start time
            end_time: ISO format end time
            description: Event description
            location: Event location
            category: Event category
            all_day: All-day event flag

        Returns:
            Success message with event ID
        """
        try:
            # Parse datetime strings
            start_dt = datetime.fromisoformat(start_time)
            end_dt = datetime.fromisoformat(end_time)

            # Create event
            event = Event(
                title=title,
                description=description,
                start_time=start_dt,
                end_time=end_dt,
                location=location,
                category=category,
                all_day=all_day,
            )

            with session_scope() as session:
                session.add(event)
                session.flush()
                event_id = event.id

            return (
                f"Created event '{title}' (ID: {event_id}) "
                f"from {start_time} to {end_time}"
            )
        except Exception as e:
            return f"Error creating event: {str(e)}"


class ListEventsInput(PydanticBaseModel):
    """Input schema for list_events tool."""

    start_date: Optional[str] = Field(
        default=None,
        description="Filter start date in ISO format (YYYY-MM-DD)",
    )
    end_date: Optional[str] = Field(
        default=None,
        description="Filter end date in ISO format (YYYY-MM-DD)",
    )
    category: Optional[str] = Field(
        default=None, description="Filter by category"
    )
    limit: int = Field(
        default=10, description="Maximum number of events to return"
    )


class ListEventsTool(BaseTool):
    """Tool for listing calendar events."""

    name: str = "list_events"
    description: str = (
        "List calendar events, optionally filtered by date range "
        "and category. Returns upcoming events by default."
    )
    args_schema: type[PydanticBaseModel] = ListEventsInput

    def _run(
        self,
        start_date: Optional[str] = None,
        end_date: Optional[str] = None,
        category: Optional[str] = None,
        limit: int = 10,
    ) -> str:
        """List calendar events.

        Args:
            start_date: Filter by start date
            end_date: Filter by end date
            category: Filter by category
            limit: Maximum results

        Returns:
            Formatted list of events
        """
        try:
            with session_scope() as session:
                query = session.query(Event)

                # Apply filters
                if start_date:
                    start_dt = datetime.fromisoformat(start_date)
                    query = query.filter(Event.start_time >= start_dt)
                else:
                    # Default to upcoming events
                    query = query.filter(Event.start_time >= datetime.now())

                if end_date:
                    end_dt = datetime.fromisoformat(end_date)
                    query = query.filter(Event.end_time <= end_dt)

                if category:
                    query = query.filter(Event.category == category)

                # Order by start time and limit
                events = query.order_by(Event.start_time).limit(limit).all()

                if not events:
                    return "No events found matching the criteria."

                # Format results
                result_lines = [f"Found {len(events)} event(s):"]
                for event in events:
                    duration = event.duration_minutes or 0
                    result_lines.append(
                        f"- [{event.id}] {event.title} | "
                        f"{event.start_time.strftime('%Y-%m-%d %H:%M')} | "
                        f"{duration}min"
                        f"{' | ' + event.location if event.location else ''}"
                        f"{' | ' + event.category if event.category else ''}"
                    )

                return "\n".join(result_lines)
        except Exception as e:
            return f"Error listing events: {str(e)}"


class UpdateEventInput(PydanticBaseModel):
    """Input schema for update_event tool."""

    event_id: int = Field(description="ID of event to update")
    title: Optional[str] = Field(default=None, description="New title")
    description: Optional[str] = Field(
        default=None, description="New description"
    )
    start_time: Optional[str] = Field(
        default=None, description="New start time (ISO format)"
    )
    end_time: Optional[str] = Field(
        default=None, description="New end time (ISO format)"
    )
    location: Optional[str] = Field(default=None, description="New location")
    category: Optional[str] = Field(default=None, description="New category")


class UpdateEventTool(BaseTool):
    """Tool for updating calendar events."""

    name: str = "update_event"
    description: str = (
        "Update an existing calendar event by ID. "
        "Only provided fields will be updated."
    )
    args_schema: type[PydanticBaseModel] = UpdateEventInput

    def _run(
        self,
        event_id: int,
        title: Optional[str] = None,
        description: Optional[str] = None,
        start_time: Optional[str] = None,
        end_time: Optional[str] = None,
        location: Optional[str] = None,
        category: Optional[str] = None,
    ) -> str:
        """Update a calendar event.

        Args:
            event_id: Event ID to update
            title: New title
            description: New description
            start_time: New start time
            end_time: New end time
            location: New location
            category: New category

        Returns:
            Success or error message
        """
        try:
            with session_scope() as session:
                event = session.query(Event).filter_by(id=event_id).first()

                if not event:
                    return f"Event with ID {event_id} not found."

                # Update provided fields
                if title is not None:
                    event.title = title
                if description is not None:
                    event.description = description
                if start_time is not None:
                    event.start_time = datetime.fromisoformat(start_time)
                if end_time is not None:
                    event.end_time = datetime.fromisoformat(end_time)
                if location is not None:
                    event.location = location
                if category is not None:
                    event.category = category

                session.commit()

                return f"Updated event '{event.title}' (ID: {event_id})"
        except Exception as e:
            return f"Error updating event: {str(e)}"


class DeleteEventInput(PydanticBaseModel):
    """Input schema for delete_event tool."""

    event_id: int = Field(description="ID of event to delete")


class DeleteEventTool(BaseTool):
    """Tool for deleting calendar events."""

    name: str = "delete_event"
    description: str = "Delete a calendar event by ID."
    args_schema: type[PydanticBaseModel] = DeleteEventInput

    def _run(self, event_id: int) -> str:
        """Delete a calendar event.

        Args:
            event_id: Event ID to delete

        Returns:
            Success or error message
        """
        try:
            with session_scope() as session:
                event = session.query(Event).filter_by(id=event_id).first()

                if not event:
                    return f"Event with ID {event_id} not found."

                title = event.title
                session.delete(event)
                session.commit()

                return f"Deleted event '{title}' (ID: {event_id})"
        except Exception as e:
            return f"Error deleting event: {str(e)}"


class CreateReminderInput(PydanticBaseModel):
    """Input schema for create_reminder tool."""

    event_id: int = Field(description="ID of event to remind about")
    reminder_time: str = Field(
        description="Reminder time in ISO format (YYYY-MM-DDTHH:MM:SS)"
    )
    message: Optional[str] = Field(
        default=None, description="Custom reminder message"
    )
    repeat_interval_minutes: Optional[int] = Field(
        default=None, description="Repeat interval in minutes"
    )


class CreateReminderTool(BaseTool):
    """Tool for creating event reminders."""

    name: str = "create_reminder"
    description: str = (
        "Create a reminder for a calendar event. "
        "Optionally set repeat interval for recurring reminders."
    )
    args_schema: type[PydanticBaseModel] = CreateReminderInput

    def _run(
        self,
        event_id: int,
        reminder_time: str,
        message: Optional[str] = None,
        repeat_interval_minutes: Optional[int] = None,
    ) -> str:
        """Create an event reminder.

        Args:
            event_id: Event to remind about
            reminder_time: When to send reminder
            message: Custom message
            repeat_interval_minutes: Repeat interval

        Returns:
            Success or error message
        """
        try:
            # Verify event exists
            with session_scope() as session:
                event = session.query(Event).filter_by(id=event_id).first()

                if not event:
                    return f"Event with ID {event_id} not found."

                # Parse reminder time
                reminder_dt = datetime.fromisoformat(reminder_time)

                # Create reminder
                reminder = Reminder(
                    event_id=event_id,
                    reminder_time=reminder_dt,
                    message=message,
                    repeat_interval_minutes=repeat_interval_minutes,
                )

                session.add(reminder)
                session.flush()
                reminder_id = reminder.id

                repeat_info = (
                    f" (repeats every {repeat_interval_minutes}min)"
                    if repeat_interval_minutes
                    else ""
                )

                return (
                    f"Created reminder (ID: {reminder_id}) for "
                    f"event '{event.title}' at {reminder_time}{repeat_info}"
                )
        except Exception as e:
            return f"Error creating reminder: {str(e)}"


class ScheduleRecurringEventInput(PydanticBaseModel):
    """Input schema for schedule_recurring_event tool."""

    title: str = Field(description="Recurring event title")
    description: Optional[str] = Field(
        default=None, description="Event description"
    )
    recurrence_type: str = Field(
        description="Recurrence type: daily, weekly, monthly, yearly"
    )
    interval: int = Field(
        default=1, description="Interval (e.g., every 2 weeks)"
    )
    start_date: str = Field(
        description="Start date in ISO format (YYYY-MM-DDTHH:MM:SS)"
    )
    end_date: Optional[str] = Field(
        default=None, description="End date in ISO format"
    )
    days_of_week: Optional[str] = Field(
        default=None,
        description="Comma-separated days (0=Mon, 6=Sun) for weekly",
    )
    location: Optional[str] = Field(default=None, description="Location")
    category: Optional[str] = Field(default=None, description="Category")


class ScheduleRecurringEventTool(BaseTool):
    """Tool for creating recurring events."""

    name: str = "schedule_recurring_event"
    description: str = (
        "Create a recurring event pattern (daily, weekly, monthly, "
        "yearly). For weekly events, specify days_of_week as "
        "comma-separated (0=Monday, 6=Sunday)."
    )
    args_schema: type[PydanticBaseModel] = ScheduleRecurringEventInput

    def _run(
        self,
        title: str,
        recurrence_type: str,
        start_date: str,
        interval: int = 1,
        description: Optional[str] = None,
        end_date: Optional[str] = None,
        days_of_week: Optional[str] = None,
        location: Optional[str] = None,
        category: Optional[str] = None,
    ) -> str:
        """Create a recurring event pattern.

        Args:
            title: Event title
            recurrence_type: daily/weekly/monthly/yearly
            start_date: Start date
            interval: Recurrence interval
            description: Description
            end_date: Optional end date
            days_of_week: Days for weekly recurrence
            location: Location
            category: Category

        Returns:
            Success or error message
        """
        try:
            # Validate recurrence type
            valid_types = ["daily", "weekly", "monthly", "yearly"]
            if recurrence_type not in valid_types:
                return (
                    f"Invalid recurrence type '{recurrence_type}'. "
                    f"Must be one of: {', '.join(valid_types)}"
                )

            # Parse dates
            start_dt = datetime.fromisoformat(start_date)
            end_dt = datetime.fromisoformat(end_date) if end_date else None

            # Create recurring event
            recurring = RecurringEvent(
                title=title,
                description=description,
                recurrence_type=recurrence_type,
                interval=interval,
                days_of_week=days_of_week,
                start_date=start_dt,
                end_date=end_dt,
                location=location,
                category=category,
            )

            with session_scope() as session:
                session.add(recurring)
                session.flush()
                recurring_id = recurring.id

            end_info = f" until {end_date}" if end_date else ""

            return (
                f"Created {recurrence_type} recurring event "
                f"'{title}' (ID: {recurring_id}) starting "
                f"{start_date}{end_info}"
            )
        except Exception as e:
            return f"Error creating recurring event: {str(e)}"


# Tool registry
CALENDAR_TOOLS = [
    CreateEventTool(),
    ListEventsTool(),
    UpdateEventTool(),
    DeleteEventTool(),
    CreateReminderTool(),
    ScheduleRecurringEventTool(),
]



"""Canvas initialization mixin for signal registration and state setup."""

from typing import Dict, Any, Optional
from PySide6.QtCore import QTimer, QRect
from PySide6.QtGui import QImage
from airunner.utils.application.mediator_mixin import MediatorMixin
from airunner.components.application.gui.windows.main.settings_mixin import (
    SettingsMixin,
)
from airunner.enums import SignalCode
from airunner.utils.settings.get_qsettings import get_qsettings


class CanvasInitializationMixin(MediatorMixin, SettingsMixin):
    """Handles canvas initialization and signal registration.

    This mixin provides functionality for:
    - Initializing canvas state variables
    - Setting up debounce timers
    - Registering signal handlers
    - Configuring performance caches
    """

    def _initialize_canvas_state(self, canvas_type: str) -> None:
        """Initialize all canvas state variables.

        Args:
            canvas_type: Type identifier for this canvas instance.
        """
        self._initialize_basic_state(canvas_type)
        self._initialize_mouse_state()
        self._initialize_history_state()
        self._initialize_event_state()
        self._initialize_persist_timer()
        self._initialize_cache_state()
        self._initialize_viewport_state()

    def _initialize_basic_state(self, canvas_type: str) -> None:
        """Initialize basic canvas state variables."""
        self._is_erasing = None
        self._is_drawing = None
        self.canvas_type = canvas_type
        self.image_backup = None
        self._current_active_image_ref = None
        self._current_active_image_binary = None
        self.previewing_filter = False
        self.painter = None
        self._painter_target = None
        self.image: Optional[QImage] = None
        self.item = None
        self._image_initialized: bool = False
        self.last_export_path = None
        self._target_size = None
        self.settings = get_qsettings()

    def _initialize_mouse_state(self) -> None:
        """Initialize mouse and interaction state variables."""
        self.last_pos = None
        self.start_pos = None
        self.selection_start_pos = None
        self.selection_stop_pos = None
        self.do_update = False
        self.generate_image_time_in_ms = 0.5
        self.do_generate_image = False
        self.generate_image_time = 0

    def _initialize_history_state(self) -> None:
        """Initialize history management state variables."""
        self.undo_history = []
        self.redo_history = []
        self._history_transactions: Dict[int | None, Dict[str, Any]] = {}
        self._structure_history_transaction: Optional[Dict[str, Any]] = None

    def _initialize_event_state(self) -> None:
        """Initialize event handling state variables."""
        self.right_mouse_button_pressed = False
        self.handling_event = False
        self._original_item_positions = {}

    def _initialize_persist_timer(self) -> None:
        """Initialize debounce timer for image persistence."""
        self._persist_timer = QTimer()
        self._persist_timer.setSingleShot(True)
        self._persist_timer.timeout.connect(self._flush_pending_image)
        self._pending_image_binary = None
        self._pending_image_ref = None
        self._persist_delay_ms = 1000
        self._active_persist_future = None
        self._persist_generation = 0

    def _initialize_cache_state(self) -> None:
        """Initialize performance caches and feature flags."""
        self._raw_image_storage_enabled = True
        self._current_active_image_hash = None
        self._qimage_cache = None
        self._qimage_cache_size = None
        self._qimage_cache_hash = None
        self._active_grid_cache = None
        self._active_grid_cache_time = 0

    def _initialize_viewport_state(self) -> None:
        """Initialize viewport and layer rendering state."""
        self._is_user_interacting = False
        self.is_dragging = False
        self._extended_viewport_rect = QRect(-2000, -2000, 4000, 4000)
        self._layer_items = {}
        self._layers_initialized = False
        self._surface_growth_step = 128
        self._minimum_surface_size = 128

    def _register_canvas_signals(self) -> None:
        """Register all signal handlers for canvas operations."""
        self._register_image_operation_signals()
        self._register_filter_signals()
        self._register_history_signals()
        self._register_layer_signals()

    def _register_image_operation_signals(self) -> None:
        """Register signals for image operations."""
        handlers = [
            (
                SignalCode.CANVAS_COPY_IMAGE_SIGNAL,
                self.on_canvas_copy_image_signal,
            ),
            (
                SignalCode.CANVAS_CUT_IMAGE_SIGNAL,
                self.on_canvas_cut_image_signal,
            ),
            (
                SignalCode.CANVAS_ROTATE_90_CLOCKWISE_SIGNAL,
                self.on_canvas_rotate_90_clockwise_signal,
            ),
            (
                SignalCode.CANVAS_ROTATE_90_COUNTER_CLOCKWISE_SIGNAL,
                self.on_canvas_rotate_90_counterclockwise_signal,
            ),
            (
                SignalCode.CANVAS_PASTE_IMAGE_SIGNAL,
                self.on_paste_image_from_clipboard,
            ),
            (
                SignalCode.CANVAS_EXPORT_IMAGE_SIGNAL,
                self.on_export_image_signal,
            ),
            (
                SignalCode.CANVAS_IMPORT_IMAGE_SIGNAL,
                self.on_import_image_signal,
            ),
            (
                SignalCode.SEND_IMAGE_TO_CANVAS_SIGNAL,
                self.on_send_image_to_canvas_signal,
            ),
            (
                SignalCode.CANVAS_LOAD_IMAGE_FROM_PATH_SIGNAL,
                self.on_load_image_from_path,
            ),
            (
                SignalCode.ENGINE_RESPONSE_WORKER_RESPONSE_SIGNAL,
                self.on_image_generated_signal,
            ),
            (SignalCode.CANVAS_CLEAR, self.on_canvas_clear_signal),
        ]
        for signal, handler in handlers:
            self.register(signal, handler)

    def _register_filter_signals(self) -> None:
        """Register signals for filter operations."""
        handlers = [
            (
                SignalCode.CANVAS_APPLY_FILTER_SIGNAL,
                self.on_apply_filter_signal,
            ),
            (
                SignalCode.CANVAS_CANCEL_FILTER_SIGNAL,
                self.on_cancel_filter_signal,
            ),
            (
                SignalCode.CANVAS_PREVIEW_FILTER_SIGNAL,
                self.on_preview_filter_signal,
            ),
        ]
        for signal, handler in handlers:
            self.register(signal, handler)

    def _register_history_signals(self) -> None:
        """Register signals for history operations."""
        handlers = [
            (SignalCode.UNDO_SIGNAL, self.on_action_undo_signal),
            (SignalCode.REDO_SIGNAL, self.on_action_redo_signal),
            (SignalCode.HISTORY_CLEAR_SIGNAL, self.on_clear_history_signal),
        ]
        for signal, handler in handlers:
            self.register(signal, handler)

    def _register_layer_signals(self) -> None:
        """Register signals for layer operations."""
        handlers = [
            (SignalCode.MASK_LAYER_TOGGLED, self.on_mask_layer_toggled),
            (
                SignalCode.LAYER_SELECTION_CHANGED,
                self._on_layer_selection_changed,
            ),
            (
                SignalCode.APPLICATION_SETTINGS_CHANGED_SIGNAL,
                self.on_settings_changed,
            ),
            (
                SignalCode.LAYER_VISIBILITY_TOGGLED,
                self.on_layer_visibility_toggled,
            ),
            (SignalCode.LAYER_DELETED, self.on_layer_deleted),
            (SignalCode.LAYER_REORDERED, self.on_layer_reordered),
            (SignalCode.LAYERS_SHOW_SIGNAL, self.on_layers_show_signal),
            (SignalCode.LAYER_OPERATION_BEGIN, self.on_layer_operation_begin),
            (
                SignalCode.LAYER_OPERATION_COMMIT,
                self.on_layer_operation_commit,
            ),
            (
                SignalCode.LAYER_OPERATION_CANCEL,
                self.on_layer_operation_cancel,
            ),
        ]
        for signal, handler in handlers:
            self.register(signal, handler)



"""Tests for AutonomousControlTools mixin."""

import unittest
from unittest.mock import Mock, MagicMock, patch
import json

from airunner.components.llm.managers.tools.autonomous_control_tools import (
    AutonomousControlTools,
)
from airunner.components.llm.tests.base_test_case import BaseTestCase
from airunner.enums import SignalCode


class MockAutonomousControlToolsClass(AutonomousControlTools):
    """Mock class for testing AutonomousControlTools mixin."""

    def __init__(self):
        self.logger = Mock()
        self.emit_signal = Mock()
        self.rag_manager = None


class TestAutonomousControlTools(BaseTestCase):
    """Test AutonomousControlTools mixin methods."""

    target_class = MockAutonomousControlToolsClass
    public_methods = [
        "get_application_state_tool",
        "schedule_task_tool",
        "propose_action_tool",
        "request_user_input_tool",
        "log_agent_decision_tool",
        "analyze_user_behavior_tool",
        "monitor_system_health_tool",
    ]

    def setUp(self):
        """Set up test with mock autonomous control tools instance."""
        super().setUp()
        self.tools = self.obj  # Alias for backwards compatibility

    def test_get_application_state_tool_creation(self):
        """Test that get_application_state_tool creates a callable tool."""
        tool = self.tools.get_application_state_tool()
        self.assertIsNotNone(tool)
        self.assertEqual(tool.name, "get_application_state")

    def test_get_application_state_returns_json(self):
        """Test that get_application_state returns valid JSON."""
        tool = self.tools.get_application_state_tool()
        result = self.invoke_tool(tool)

        # Should be valid JSON string
        state = json.loads(result)
        self.assertIsInstance(state, dict)
        self.assertIn("application", state)
        self.assertIn("llm", state)

    def test_schedule_task_tool_creation(self):
        """Test that schedule_task_tool creates a callable tool."""
        tool = self.tools.schedule_task_tool()
        self.assertIsNotNone(tool)
        self.assertEqual(tool.name, "schedule_task")

    def test_schedule_task_immediate(self):
        """Test scheduling an immediate task."""
        tool = self.tools.schedule_task_tool()
        result = self.invoke_tool(
            tool,
            task_name="test_task",
            description="Test task description",
            when="now",
        )

        self.assertIn("Scheduled task", result)
        self.assertIn("test_task", result)

        # Verify signal was emitted
        self.tools.emit_signal.assert_called_with(
            SignalCode.SCHEDULE_TASK_SIGNAL,
            {
                "task_name": "test_task",
                "description": "Test task description",
                "when": "now",
                "params": None,
            },
        )

    def test_schedule_task_with_params(self):
        """Test scheduling task with parameters."""
        tool = self.tools.schedule_task_tool()
        params = {"image_prompt": "a cat", "steps": 30}
        result = self.invoke_tool(
            tool,
            task_name="generate_image",
            description="Generate cat image",
            when="in 5 minutes",
            params=params,
        )

        self.assertIn("Scheduled task", result)

        # Verify signal includes params
        call_args = self.tools.emit_signal.call_args
        self.assertEqual(call_args[0][0], SignalCode.SCHEDULE_TASK_SIGNAL)
        self.assertEqual(call_args[0][1]["params"], params)

    def test_set_application_mode_tool_creation(self):
        """Test that set_application_mode_tool creates a callable tool."""
        tool = self.tools.set_application_mode_tool()
        self.assertIsNotNone(tool)
        self.assertEqual(tool.name, "set_application_mode")

    def test_set_application_mode_autonomous(self):
        """Test setting application to autonomous mode."""
        tool = self.tools.set_application_mode_tool()
        result = self.invoke_tool(
            tool, mode="autonomous", reason="Testing autonomous mode"
        )

        self.assertIn("Set application mode", result)
        self.assertIn("autonomous", result)

        # Verify signal was emitted
        self.tools.emit_signal.assert_called_with(
            SignalCode.SET_APPLICATION_MODE_SIGNAL,
            {"mode": "autonomous", "reason": "Testing autonomous mode"},
        )

    def test_set_application_mode_validation(self):
        """Test that invalid modes are rejected."""
        tool = self.tools.set_application_mode_tool()
        result = self.invoke_tool(tool, mode="invalid_mode", reason="Test")

        self.assertIn("must be one of", result.lower())

    def test_request_user_input_tool_creation(self):
        """Test that request_user_input_tool creates a callable tool."""
        tool = self.tools.request_user_input_tool()
        self.assertIsNotNone(tool)
        self.assertEqual(tool.name, "request_user_input")

    def test_request_user_input_approval(self):
        """Test requesting user approval."""
        tool = self.tools.request_user_input_tool()
        result = self.invoke_tool(
            tool,
            prompt="Delete old files?",
            input_type="approval",
            context={"file_count": 42},
        )

        self.assertIn("Requested user input", result)

        # Verify signal was emitted
        self.tools.emit_signal.assert_called()
        call_args = self.tools.emit_signal.call_args
        self.assertEqual(call_args[0][0], SignalCode.REQUEST_USER_INPUT_SIGNAL)
        self.assertIn("prompt", call_args[0][1])

    def test_analyze_user_behavior_tool_creation(self):
        """Test that analyze_user_behavior_tool creates a callable tool."""
        tool = self.tools.analyze_user_behavior_tool()
        self.assertIsNotNone(tool)
        self.assertEqual(tool.name, "analyze_user_behavior")

    def test_propose_action_tool_creation(self):
        """Test that propose_action_tool creates a callable tool."""
        tool = self.tools.propose_action_tool()
        self.assertIsNotNone(tool)
        self.assertEqual(tool.name, "propose_action")

    def test_propose_action_with_rationale(self):
        """Test proposing an action with rationale."""
        tool = self.tools.propose_action_tool()
        result = self.invoke_tool(
            tool,
            action="cleanup_cache",
            rationale="Cache is taking up 5GB of disk space",
            confidence=0.85,
            requires_approval=True,
        )

        self.assertIn("Proposed action", result)
        self.assertIn("cleanup_cache", result)

        # Verify signal was emitted
        self.tools.emit_signal.assert_called_with(
            SignalCode.AGENT_ACTION_PROPOSAL_SIGNAL,
            {
                "action": "cleanup_cache",
                "rationale": "Cache is taking up 5GB of disk space",
                "confidence": 0.85,
                "requires_approval": True,
            },
        )

    @patch("psutil.cpu_percent", return_value=45.0)
    @patch("psutil.virtual_memory")
    @patch("psutil.disk_usage")
    def test_monitor_system_health_tool_creation(
        self, mock_disk, mock_mem, mock_cpu
    ):
        """Test that monitor_system_health_tool creates a callable tool."""
        # Mock memory
        mock_mem.return_value = MagicMock(percent=60.0)
        # Mock disk
        mock_disk.return_value = MagicMock(percent=70.0)

        tool = self.tools.monitor_system_health_tool()
        self.assertIsNotNone(tool)
        self.assertEqual(tool.name, "monitor_system_health")

    @patch("psutil.cpu_percent", return_value=45.0)
    @patch("psutil.virtual_memory")
    @patch("psutil.disk_usage")
    def test_monitor_system_health_returns_metrics(
        self, mock_disk, mock_mem, mock_cpu
    ):
        """Test that monitor_system_health returns valid metrics."""
        # Mock memory
        mock_mem.return_value = MagicMock(percent=60.0)
        # Mock disk
        mock_disk.return_value = MagicMock(percent=70.0)

        tool = self.tools.monitor_system_health_tool()
        result = self.invoke_tool(tool)

        self.assertIn("CPU usage", result)
        self.assertIn("45.0%", result)
        self.assertIn("Memory usage", result)
        self.assertIn("60.0%", result)

    def test_log_agent_decision_tool_creation(self):
        """Test that log_agent_decision_tool creates a callable tool."""
        tool = self.tools.log_agent_decision_tool()
        self.assertIsNotNone(tool)
        self.assertEqual(tool.name, "log_agent_decision")

    def test_log_agent_decision_with_context(self):
        """Test logging agent decision with context."""
        tool = self.tools.log_agent_decision_tool()
        context = {"user_request": "summarize document", "token_count": 5000}
        result = self.invoke_tool(
            tool,
            decision="split_into_chunks",
            reasoning="Document too large for single pass",
            context=context,
        )

        self.assertIn("Logged agent decision", result)
        self.assertIn("split_into_chunks", result)


if __name__ == "__main__":
    unittest.main()



"""Mixins for dataset manager functionality."""

from .cache_mixin import DatasetCacheMixin
from .download_mixin import DatasetDownloadMixin
from .loader_mixin import DatasetLoaderMixin

__all__ = [
    "DatasetCacheMixin",
    "DatasetDownloadMixin",
    "DatasetLoaderMixin",
]



"""
zimreader.py

Provides ZIMReader for querying ZIM archives using libzim.

"""

from typing import Optional, List
from libzim.reader import Archive, Entry
from libzim.search import Query, Searcher

from airunner.settings import AIRUNNER_LOG_LEVEL
from airunner.utils.application import get_logger

logger = get_logger(__name__, AIRUNNER_LOG_LEVEL)


class ZIMReader:
    """ZIMReader provides search and retrieval for ZIM archives using libzim.

    Args:
        zim_file_path (str): Path to the ZIM archive file.
    """

    def __init__(self, zim_file_path: str) -> None:
        self.zim_file_path = zim_file_path
        try:
            self.archive = Archive(zim_file_path)
            # Log available methods for debugging
            logger.debug(
                f"Archive methods: {[m for m in dir(self.archive) if not m.startswith('_')]}"
            )
        except Exception as e:
            logger.error(f"Failed to open ZIM archive: {zim_file_path}: {e}")
            raise

    def get_article(self, path: str) -> Optional[str]:
        """Retrieve article HTML content by canonical path.

        Args:
            path (str): Canonical ZIM path (e.g., 'A/Albert_Einstein.html').

        Returns:
            Optional[str]: HTML content if found, else None.
        """
        try:
            entry: Entry = self.archive.get_entry_by_path(path)
            content_bytes = entry.get_item().content
            return content_bytes.tobytes().decode("utf-8")
        except Exception as e:
            logger.warning(f"Article not found or failed to read: {path}: {e}")
            return None

    def get_all_entry_paths(self, limit: Optional[int] = None) -> List[str]:
        """Get entry paths from the ZIM archive by random sampling.

        Since libzim's Python bindings don't support iteration, we use
        get_random_entry() to sample entries from the archive.

        Args:
            limit (Optional[int]): Maximum number of entries to return. Default 1000.

        Returns:
            List[str]: List of canonical paths for entries.
        """
        paths = []
        seen_paths = set()

        try:
            # Get entry count to know archive size
            entry_count = self.archive.entry_count
            logger.info(f"Archive has {entry_count} total entries")

            # Set reasonable limit
            max_entries = limit if limit else min(1000, entry_count)

            # Use get_random_entry() to sample entries
            # Sample more than needed since we'll filter some out
            attempts = (
                max_entries * 3
            )  # Try 3x to account for redirects/metadata

            for _ in range(attempts):
                if len(paths) >= max_entries:
                    break

                try:
                    entry = self.archive.get_random_entry()

                    # Skip redirects and metadata
                    if entry.is_redirect:
                        continue

                    path = entry.path

                    # Filter metadata entries and avoid duplicates
                    if not path.startswith("-/") and path not in seen_paths:
                        paths.append(path)
                        seen_paths.add(path)

                except Exception as e:
                    logger.debug(f"Error getting random entry: {e}")
                    continue

            logger.info(
                f"Successfully retrieved {len(paths)} unique entry paths"
            )

        except Exception as e:
            logger.error(f"Failed to enumerate entries: {e}")

        return paths

    def search(self, query: str, limit: int = 5) -> List[str]:
        """Full-text search for articles in the ZIM archive.

        Args:
            query (str): Search query string.
            limit (int): Maximum number of results.

        Returns:
            List[str]: List of canonical paths for matching articles.
        """
        try:
            searcher = Searcher(self.archive)
            # Modern libzim API: Query constructor takes the query string directly
            q = Query(query) if query else Query("")
            search = searcher.search(q)
            return [result.path for result in search.getResults(0, limit)]
        except Exception as e:
            logger.error(f"Search failed for query '{query}': {e}")
            return []



"""Training utilities for diffusion models.

This package provides trainer classes that can be used from the command line
or imported and executed in a background thread inside the GUI.

Modules:
- base: Common config, base class, logging helpers.
- datasets: Simple image+caption dataset loaders for folder-based data.
- finetune: Full or UNet-only fine-tuning trainer.
- textual_inversion: Trainer for learnable token embeddings.
"""

from airunner.components.art.trainers.base import TrainingConfig, BaseTrainer
from airunner.components.art.trainers.finetune import SDTextToImageTrainer
from airunner.components.art.trainers.textual_inversion import (
    TextualInversionTrainer,
)

__all__ = [
    "TrainingConfig",
    "BaseTrainer",
    "SDTextToImageTrainer",
    "TextualInversionTrainer",
]



"""
Properties mixin for X4UpscaleManager.

This mixin contains all property accessors for the X4 upscale manager including
compel settings, preview paths, pipeline configuration, and loading status.
"""

import os
from typing import Any, Dict

from diffusers import StableDiffusionUpscalePipeline

from airunner.enums import ModelStatus


class X4PropertiesMixin:
    """Properties for X4UpscaleManager configuration and state."""

    @property
    def use_compel(self) -> bool:
        """X4 upscaler does not support Compel prompt weighting.

        Returns:
            Always False for upscaler pipeline.
        """
        return False

    @property
    def preview_dir(self) -> str:
        """Directory for storing upscaled preview images.

        Returns:
            Absolute path to upscaled images directory.
        """
        return os.path.join(
            os.path.expanduser(self.path_settings.base_path),
            "art",
            "upscaled",
        )

    @property
    def preview_path(self) -> str:
        """Path to current upscale preview image.

        Returns:
            Full path to preview file.
        """
        return os.path.join(self.preview_dir, self.PREVIEW_FILENAME)

    @property
    def use_from_single_file(self) -> bool:
        """X4 upscaler uses standard diffusers format, not single file.

        Returns:
            Always False for upscaler pipeline.
        """
        return False

    @property
    def pipeline_map(self) -> Dict[str, Any]:
        """Mapping of pipeline names to pipeline classes.

        Returns:
            Dictionary mapping 'x4-upscaler' to StableDiffusionUpscalePipeline.
        """
        return {"x4-upscaler": StableDiffusionUpscalePipeline}

    def is_loaded(self) -> bool:
        """Check if upscaler pipeline is loaded and ready.

        Returns:
            True if pipeline is loaded and status is LOADED.
        """
        return (
            self._pipe is not None and self.model_status == ModelStatus.LOADED
        )



"""
General utility functions for Stable Diffusion handlers.
Handles image resizing, callbacks, seed setting, and other helpers.
Follows project standards: docstrings, type hints, logging.
"""

from typing import Optional
from PIL.Image import Image
import PIL

from airunner.settings import AIRUNNER_LOG_LEVEL
from airunner.utils.application import get_logger

logger = get_logger(__name__, AIRUNNER_LOG_LEVEL)


def resize_image(
    image: Image, max_width: int, max_height: int
) -> Optional[Image]:
    """Resize the image to fit within max_width and max_height, maintaining aspect ratio."""
    if image is None:
        return None
    original_width, original_height = image.size
    if original_width <= max_width and original_height <= max_height:
        return image
    aspect_ratio = original_width / original_height
    if aspect_ratio > 1:
        new_width = min(max_width, original_width)
        new_height = int(new_width / aspect_ratio)
    else:
        new_height = min(max_height, original_height)
        new_width = int(new_height * aspect_ratio)
    resized_image = image.resize(
        (new_width, new_height), PIL.Image.Resampling.LANCZOS
    )
    logger.info(f"Resized image to {new_width}x{new_height}")
    return resized_image


# Additional general utilities (callbacks, seed setting, etc.) can be added here.



import types


from airunner.components.art.managers.stablediffusion import model_loader


def test_load_embedding_path_missing(monkeypatch, tmp_path):
    emb = types.SimpleNamespace(path=str(tmp_path / "nope.pt"), name="e1")

    class DummyPipe:
        pass

    res = model_loader.load_embedding(path=emb.path)
    # current stub returns an object with the path attribute
    assert hasattr(res, "path")
    assert res.path == emb.path


def test_load_lora_weights_success(monkeypatch):
    class DummyPipe:
        def load_lora_weights(self, base, weight_name=None, adapter_name=None):
            return None

    lora = types.SimpleNamespace(path="/some/path/adapter.safetensors")
    res = model_loader.load_lora_weights(
        DummyPipe(),
        lora,
        "/some/path",
        logger=types.SimpleNamespace(
            info=lambda *a, **k: None, warning=lambda *a, **k: None
        ),
    )
    assert res is True


def test_unload_functions_noop(monkeypatch):
    class Pipe:
        pass

    pipe = Pipe()
    # functions should not raise
    model_loader.unload_lora(
        pipe,
        logger=types.SimpleNamespace(
            info=lambda *a, **k: None, warning=lambda *a, **k: None
        ),
    )
    model_loader.unload_embeddings(
        pipe,
        logger=types.SimpleNamespace(
            info=lambda *a, **k: None, warning=lambda *a, **k: None
        ),
    )


def test_load_compel_proc_and_unload(monkeypatch):
    class C:
        def __init__(self, **kwargs):
            pass

    monkeypatch.setitem(
        __import__("sys").modules, "compel", types.SimpleNamespace(Compel=C)
    )
    proc = model_loader.load_compel_proc(
        {},
        logger=types.SimpleNamespace(
            info=lambda *a, **k: None, error=lambda *a, **k: None
        ),
    )
    assert proc is not None
    # unload should not raise
    model_loader.unload_compel_proc(
        proc,
        logger=types.SimpleNamespace(
            info=lambda *a, **k: None, warning=lambda *a, **k: None
        ),
    )


def test_load_deep_cache_helper_and_unload(monkeypatch):
    class Helper:
        def __init__(self, pipe=None):
            pass

        def set_params(self, **kw):
            pass

        def enable(self):
            pass

        def disable(self):
            pass

    monkeypatch.setitem(
        __import__("sys").modules,
        "DeepCache",
        types.SimpleNamespace(DeepCacheSDHelper=Helper),
    )
    helper = model_loader.load_deep_cache_helper(
        pipe=object(),
        logger=types.SimpleNamespace(
            info=lambda *a, **k: None, error=lambda *a, **k: None
        ),
    )
    assert helper is not None
    model_loader.unload_deep_cache_helper(
        helper,
        logger=types.SimpleNamespace(
            info=lambda *a, **k: None, warning=lambda *a, **k: None
        ),
    )


def test_load_controlnet_processor_returns_none_when_disabled():
    res = model_loader.load_controlnet_processor(
        False,
        controlnet_model=None,
        controlnet_processor_path="/tmp",
        logger=types.SimpleNamespace(
            info=lambda *a, **k: None, error=lambda *a, **k: None
        ),
    )
    assert res is None



"""Tests for workflow template manager."""

import pytest
import tempfile
from pathlib import Path

from airunner.components.nodegraph.template_manager import (
    TemplateManager,
    WorkflowTemplate,
    get_template_manager,
)


class TestWorkflowTemplate:
    """Test WorkflowTemplate class."""

    def test_create_template(self):
        """Test creating template from dictionary."""
        data = {
            "name": "Test Template",
            "description": "A test template",
            "category": "test",
            "tags": ["test", "demo"],
            "author": "Tester",
            "version": "1.0",
            "variables": {"var1": "value1"},
            "nodes": [{"node_identifier": "test.Node", "name": "n1"}],
            "connections": [],
        }

        template = WorkflowTemplate(data)

        assert template.name == "Test Template"
        assert template.description == "A test template"
        assert template.category == "test"
        assert template.tags == ["test", "demo"]
        assert template.author == "Tester"
        assert template.version == "1.0"
        assert template.variables == {"var1": "value1"}
        assert len(template.nodes) == 1
        assert len(template.connections) == 0

    def test_template_defaults(self):
        """Test template with missing fields uses defaults."""
        data = {"nodes": [], "connections": []}

        template = WorkflowTemplate(data)

        assert template.name == "Untitled Template"
        assert template.description == ""
        assert template.category == "general"
        assert template.tags == []
        assert template.author == "Unknown"
        assert template.version == "1.0"

    def test_to_dict(self):
        """Test converting template to dictionary."""
        data = {
            "name": "Test",
            "description": "Test template",
            "category": "test",
            "tags": ["test"],
            "author": "Tester",
            "version": "1.0",
            "variables": {},
            "nodes": [],
            "connections": [],
        }

        template = WorkflowTemplate(data)
        result = template.to_dict()

        assert result["name"] == "Test"
        assert result["description"] == "Test template"
        assert result["category"] == "test"


class TestTemplateManager:
    """Test TemplateManager class."""

    @pytest.fixture
    def temp_dir(self):
        """Create temporary directory for templates."""
        with tempfile.TemporaryDirectory() as tmpdir:
            yield Path(tmpdir)

    @pytest.fixture
    def manager(self, temp_dir):
        """Create template manager with temp directory."""
        return TemplateManager(str(temp_dir))

    def test_manager_initialization(self, temp_dir):
        """Test manager initialization creates directory."""
        assert temp_dir.exists()

        manager = TemplateManager(str(temp_dir))
        assert manager.template_dir == temp_dir
        assert len(manager.templates) == 0
        assert len(manager.categories) == 0

    def test_save_and_load_template(self, manager, temp_dir):
        """Test saving and loading templates."""
        data = {
            "name": "Test Template",
            "description": "A test",
            "category": "test",
            "tags": [],
            "author": "Test",
            "version": "1.0",
            "variables": {},
            "nodes": [],
            "connections": [],
        }

        template = WorkflowTemplate(data)
        success = manager.save_template(template)

        assert success is True
        assert "Test Template" in manager.templates

        # Check file was created
        files = list(temp_dir.glob("*.json"))
        assert len(files) == 1

        # Reload and check
        manager.reload()
        loaded = manager.get_template("Test Template")
        assert loaded is not None
        assert loaded.name == "Test Template"

    def test_list_templates(self, manager):
        """Test listing templates."""
        # Create multiple templates
        for i in range(3):
            data = {
                "name": f"Template {i}",
                "description": "Test",
                "category": "test",
                "nodes": [],
                "connections": [],
            }
            template = WorkflowTemplate(data)
            manager.save_template(template)

        templates = manager.list_templates()
        assert len(templates) == 3

    def test_list_by_category(self, manager):
        """Test listing templates by category."""
        # Create templates in different categories
        categories = ["cat1", "cat2", "cat1"]
        for i, cat in enumerate(categories):
            data = {
                "name": f"Template {i}",
                "description": "Test",
                "category": cat,
                "nodes": [],
                "connections": [],
            }
            template = WorkflowTemplate(data)
            manager.save_template(template)

        cat1_templates = manager.list_by_category("cat1")
        cat2_templates = manager.list_by_category("cat2")

        assert len(cat1_templates) == 2
        assert len(cat2_templates) == 1

    def test_list_categories(self, manager):
        """Test listing all categories."""
        # Create templates in different categories
        for cat in ["cat1", "cat2", "cat3"]:
            data = {
                "name": f"Template {cat}",
                "description": "Test",
                "category": cat,
                "nodes": [],
                "connections": [],
            }
            template = WorkflowTemplate(data)
            manager.save_template(template)

        categories = manager.list_categories()
        assert len(categories) == 3
        assert "cat1" in categories
        assert "cat2" in categories
        assert "cat3" in categories

    def test_search_templates_by_name(self, manager):
        """Test searching templates by name."""
        names = ["RAG Workflow", "Agent Loop", "Tool Chain"]
        for name in names:
            data = {
                "name": name,
                "description": "Test",
                "category": "test",
                "nodes": [],
                "connections": [],
            }
            template = WorkflowTemplate(data)
            manager.save_template(template)

        results = manager.search_templates("RAG")
        assert len(results) == 1
        assert results[0].name == "RAG Workflow"

    def test_search_templates_by_description(self, manager):
        """Test searching templates by description."""
        data = {
            "name": "Test",
            "description": "This is a special workflow template",
            "category": "test",
            "nodes": [],
            "connections": [],
        }
        template = WorkflowTemplate(data)
        manager.save_template(template)

        results = manager.search_templates("special")
        assert len(results) == 1

    def test_search_templates_by_tag(self, manager):
        """Test searching templates by tags."""
        data = {
            "name": "Test",
            "description": "Test",
            "category": "test",
            "tags": ["rag", "knowledge", "search"],
            "nodes": [],
            "connections": [],
        }
        template = WorkflowTemplate(data)
        manager.save_template(template)

        results = manager.search_templates("knowledge")
        assert len(results) == 1

    def test_search_with_category_filter(self, manager):
        """Test searching with category filter."""
        # Create templates in different categories
        data1 = {
            "name": "RAG Workflow",
            "description": "Test",
            "category": "knowledge",
            "nodes": [],
            "connections": [],
        }
        data2 = {
            "name": "RAG Agent",
            "description": "Test",
            "category": "agent",
            "nodes": [],
            "connections": [],
        }

        manager.save_template(WorkflowTemplate(data1))
        manager.save_template(WorkflowTemplate(data2))

        # Search for "RAG" in knowledge category only
        results = manager.search_templates("RAG", category="knowledge")
        assert len(results) == 1
        assert results[0].category == "knowledge"

    def test_delete_template(self, manager, temp_dir):
        """Test deleting template."""
        data = {
            "name": "To Delete",
            "description": "Test",
            "category": "test",
            "nodes": [],
            "connections": [],
        }
        template = WorkflowTemplate(data)
        manager.save_template(template)

        # Verify it exists
        assert manager.get_template("To Delete") is not None

        # Delete it
        success = manager.delete_template("To Delete")
        assert success is True

        # Verify it's gone
        assert manager.get_template("To Delete") is None

        # Verify file is deleted
        files = list(temp_dir.glob("*.json"))
        assert len(files) == 0

    def test_create_from_template_basic(self, manager):
        """Test creating workflow from template."""
        data = {
            "name": "Test Template",
            "description": "Test",
            "category": "test",
            "variables": {"var1": "default1"},
            "nodes": [
                {
                    "node_identifier": "test.Node",
                    "name": "node1",
                    "properties": {"value": "{{var1}}"},
                }
            ],
            "connections": [],
        }
        template = WorkflowTemplate(data)
        manager.save_template(template)

        workflow = manager.create_from_template("Test Template", "My Workflow")

        assert workflow is not None
        assert workflow["name"] == "My Workflow"
        assert workflow["variables"]["var1"] == "default1"
        assert workflow["nodes"][0]["properties"]["value"] == "default1"

    def test_create_from_template_with_variable_override(self, manager):
        """Test creating workflow with variable overrides."""
        data = {
            "name": "Test Template",
            "description": "Test",
            "category": "test",
            "variables": {"var1": "default1", "var2": "default2"},
            "nodes": [
                {
                    "node_identifier": "test.Node",
                    "name": "node1",
                    "properties": {
                        "value1": "{{var1}}",
                        "value2": "{{var2}}",
                    },
                }
            ],
            "connections": [],
        }
        template = WorkflowTemplate(data)
        manager.save_template(template)

        workflow = manager.create_from_template(
            "Test Template",
            "My Workflow",
            variables={"var1": "override1"},
        )

        assert workflow is not None
        assert workflow["variables"]["var1"] == "override1"
        assert workflow["variables"]["var2"] == "default2"
        assert workflow["nodes"][0]["properties"]["value1"] == "override1"
        assert workflow["nodes"][0]["properties"]["value2"] == "default2"

    def test_create_from_nonexistent_template(self, manager):
        """Test creating from non-existent template returns None."""
        workflow = manager.create_from_template("Nonexistent", "My Workflow")
        assert workflow is None

    def test_load_invalid_json(self, temp_dir):
        """Test loading invalid JSON file."""
        # Create invalid JSON file
        invalid_file = temp_dir / "invalid.json"
        invalid_file.write_text("{ invalid json }")

        manager = TemplateManager(str(temp_dir))

        # Should not crash, just log error and skip file
        assert len(manager.templates) == 0

    def test_get_template_manager_singleton(self):
        """Test get_template_manager returns singleton."""
        manager1 = get_template_manager()
        manager2 = get_template_manager()

        assert manager1 is manager2



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'user_settings.ui'
##
## Created by: Qt User Interface Compiler version 6.9.0
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QApplication, QComboBox, QGridLayout, QGroupBox,
    QLineEdit, QSizePolicy, QSpacerItem, QWidget)

class Ui_user_settings_widget(object):
    def setupUi(self, user_settings_widget):
        if not user_settings_widget.objectName():
            user_settings_widget.setObjectName(u"user_settings_widget")
        user_settings_widget.resize(500, 346)
        self.gridLayout = QGridLayout(user_settings_widget)
        self.gridLayout.setObjectName(u"gridLayout")
        self.groupBox_3 = QGroupBox(user_settings_widget)
        self.groupBox_3.setObjectName(u"groupBox_3")
        self.gridLayout_2 = QGridLayout(self.groupBox_3)
        self.gridLayout_2.setObjectName(u"gridLayout_2")
        self.unit_system = QComboBox(self.groupBox_3)
        self.unit_system.addItem("")
        self.unit_system.addItem("")
        self.unit_system.setObjectName(u"unit_system")

        self.gridLayout_2.addWidget(self.unit_system, 0, 0, 1, 1)


        self.gridLayout.addWidget(self.groupBox_3, 2, 0, 1, 1)

        self.groupBox_2 = QGroupBox(user_settings_widget)
        self.groupBox_2.setObjectName(u"groupBox_2")
        self.gridLayout_4 = QGridLayout(self.groupBox_2)
        self.gridLayout_4.setObjectName(u"gridLayout_4")
        self.zipcode = QLineEdit(self.groupBox_2)
        self.zipcode.setObjectName(u"zipcode")

        self.gridLayout_4.addWidget(self.zipcode, 0, 0, 1, 1)


        self.gridLayout.addWidget(self.groupBox_2, 1, 0, 1, 1)

        self.groupBox = QGroupBox(user_settings_widget)
        self.groupBox.setObjectName(u"groupBox")
        self.gridLayout_3 = QGridLayout(self.groupBox)
        self.gridLayout_3.setObjectName(u"gridLayout_3")
        self.username = QLineEdit(self.groupBox)
        self.username.setObjectName(u"username")

        self.gridLayout_3.addWidget(self.username, 0, 0, 1, 1)


        self.gridLayout.addWidget(self.groupBox, 0, 0, 1, 1)

        self.verticalSpacer = QSpacerItem(20, 40, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)

        self.gridLayout.addItem(self.verticalSpacer, 3, 0, 1, 1)


        self.retranslateUi(user_settings_widget)
        self.username.textChanged.connect(user_settings_widget.username_changed)
        self.zipcode.textChanged.connect(user_settings_widget.zipcode_changed)
        self.unit_system.currentTextChanged.connect(user_settings_widget.unit_system_changed)

        QMetaObject.connectSlotsByName(user_settings_widget)
    # setupUi

    def retranslateUi(self, user_settings_widget):
        user_settings_widget.setWindowTitle(QCoreApplication.translate("user_settings_widget", u"Form", None))
        self.groupBox_3.setTitle(QCoreApplication.translate("user_settings_widget", u"Unit System", None))
        self.unit_system.setItemText(0, QCoreApplication.translate("user_settings_widget", u"Imperial", None))
        self.unit_system.setItemText(1, QCoreApplication.translate("user_settings_widget", u"Metric", None))

        self.groupBox_2.setTitle(QCoreApplication.translate("user_settings_widget", u"Zipcode", None))
        self.groupBox.setTitle(QCoreApplication.translate("user_settings_widget", u"Username", None))
    # retranslateUi




#!/usr/bin/env python
"""
Test runner for LLM component tests.

Usage:
    python -m airunner.components.llm.tests.run_tests

Or run specific test:
    python -m airunner.components.llm.tests.test_tool_manager
"""
import unittest
import sys
from pathlib import Path


def run_tests():
    """Discover and run all tests in the llm/tests directory."""
    # Get the tests directory
    tests_dir = Path(__file__).parent

    # Discover tests
    loader = unittest.TestLoader()
    suite = loader.discover(
        start_dir=str(tests_dir),
        pattern="test_*.py",
        top_level_dir=str(tests_dir.parent.parent.parent),
    )

    # Run tests
    runner = unittest.TextTestRunner(verbosity=2)
    result = runner.run(suite)

    # Exit with error code if tests failed
    sys.exit(0 if result.wasSuccessful() else 1)


if __name__ == "__main__":
    run_tests()



from PySide6.QtWidgets import (
    QDialog,
    QVBoxLayout,
    QListWidget,
    QListWidgetItem,
    QDialogButtonBox,
    QWidget,
    QApplication,
    QCheckBox,
    QHBoxLayout,
)
from PySide6.QtCore import Qt

from airunner.components.art.data.image_filter import ImageFilter
from airunner.settings import AIRUNNER_LOG_LEVEL
from airunner.utils.application import get_logger


class FilterListWindow(QDialog):
    """Simple dialog listing available image filters with checkboxes to toggle
    auto-apply. Checking a filter updates the `auto_apply` column in the DB
    immediately.
    """

    def __init__(self, parent=None):
        # Initialize QDialog with an optional parent and show it non-modally
        super().__init__(parent)
        self.logger = get_logger(__name__, AIRUNNER_LOG_LEVEL)
        self.setWindowTitle("Image Filters")
        # Provide a sensible default size so the window isn't tiny
        try:
            self.resize(420, 480)
            self.setMinimumSize(300, 240)
        except Exception:
            pass
        self._list = QListWidget(self)
        self._list.setSelectionMode(QListWidget.NoSelection)
        # Inherit the application's stylesheet (if any) so dialog visuals
        # match the rest of the app. Fall back to parent's stylesheet first
        # then to the QApplication stylesheet.
        try:
            app = QApplication.instance()
            parent_ss = None
            if parent is not None and hasattr(parent, "styleSheet"):
                parent_ss = parent.styleSheet()
            if parent_ss:
                self.setStyleSheet(parent_ss)
            elif app is not None and app.styleSheet():
                self.setStyleSheet(app.styleSheet())
        except Exception:
            pass

        # Styling is provided by the app theme QSS (inherited above).

        self._buttons = QDialogButtonBox(
            QDialogButtonBox.Ok | QDialogButtonBox.Cancel
        )
        self._buttons.accepted.connect(self.accept)
        self._buttons.rejected.connect(self.reject)

        layout = QVBoxLayout(self)
        layout.addWidget(self._list)
        layout.addWidget(self._buttons)

        self._load_filters()
        # Show non-modally and raise so it appears on top
        self.show()
        try:
            self.raise_()
        except Exception:
            pass

    def _load_filters(self):
        self._list.clear()
        filters = ImageFilter.objects.all() or []
        for f in filters:
            # Create an empty list item and a widget containing a real QCheckBox
            item = QListWidgetItem()
            container = QWidget()
            row_layout = QHBoxLayout(container)
            row_layout.setContentsMargins(4, 2, 4, 2)
            row_layout.setSpacing(8)

            checkbox = QCheckBox(f.display_name or f.name, container)
            checkbox.setChecked(bool(f.auto_apply))

            # Connect directly to update DB when toggled so we don't rely on
            # QListWidget's itemChanged which may not trigger QCheckBox styling.
            def _on_checkbox_toggled(checked: bool, fid=f.id):
                try:
                    self.logger.debug(
                        "Updating ImageFilter id=%s auto_apply=%s",
                        fid,
                        checked,
                    )
                    success = ImageFilter.objects.update(
                        fid, auto_apply=bool(checked)
                    )
                    self.logger.debug("ImageFilter update success=%s", success)
                except Exception:
                    self.logger.exception(
                        "Failed to update ImageFilter(id=%s) auto_apply", fid
                    )

            # Use toggled(bool) for a clear boolean value and to avoid any
            # tri-state or integer/enum mismatch that can occur with
            # stateChanged(int).
            checkbox.toggled.connect(_on_checkbox_toggled)

            row_layout.addWidget(checkbox)
            row_layout.addStretch(1)

            item.setSizeHint(container.sizeHint())
            # store id for later if needed
            item.setData(Qt.UserRole, f.id)
            self._list.addItem(item)
            self._list.setItemWidget(item, container)

        # We use embedded QCheckBox widgets and connect their stateChanged
        # signals directly to the DB update. Avoid connecting the legacy
        # QListWidget.itemChanged handler because it reads the
        # QListWidgetItem.checkState() and can overwrite updates when the
        # list item isn't actually used as the checkbox owner.

    def _on_item_changed(self, item: QListWidgetItem):
        try:
            fid = item.data(Qt.UserRole)
            checked = item.checkState() == Qt.Checked
            ImageFilter.objects.update(fid, auto_apply=checked)
            # Do NOT apply to current canvas image when toggled; these flags
            # are for auto-applying to future generated images. The DB update
            # above is sufficient.
        except Exception:
            self.logger.exception(
                "Error handling itemChanged for ImageFilter id=%s", fid
            )

    def accept(self):
        super().accept()

    def reject(self):
        super().reject()



#!/usr/bin/env python3
"""
Quick test to verify scheduler switching behavior.
Run this to check if schedulers are actually being changed correctly.
"""
import diffusers
from airunner.components.art.data.schedulers import Schedulers
from airunner.enums import Scheduler

# Test that our schedulers database has the right mappings
print("Testing scheduler mappings...")
print("-" * 60)

test_schedulers = [
    Scheduler.EULER.value,
    Scheduler.DPM.value,
    Scheduler.DPM_PP_2M.value,
    Scheduler.DPM_PP_2M_K.value,
    Scheduler.DPM_PP_2M_SDE_K.value,
]

for scheduler_name in test_schedulers:
    scheduler_record = Schedulers.objects.filter_by_first(
        display_name=scheduler_name
    )
    if scheduler_record:
        scheduler_class_name = scheduler_record.name
        scheduler_class = getattr(diffusers, scheduler_class_name, None)
        if scheduler_class:
            print(f"✓ {scheduler_name:25s} -> {scheduler_class_name}")
        else:
            print(
                f"✗ {scheduler_name:25s} -> {scheduler_class_name} (CLASS NOT FOUND)"
            )
    else:
        print(f"✗ {scheduler_name:25s} -> NOT IN DATABASE")

print("\n" + "=" * 60)
print("Testing DPM scheduler config overrides...")
print("-" * 60)

# Create a dummy config to test the override logic
from airunner.enums import Scheduler

test_cases = [
    (
        Scheduler.DPM,
        {"algorithm_type": "dpmsolver", "use_karras_sigmas": False},
    ),
    (
        Scheduler.DPM_PP_2M,
        {"algorithm_type": "dpmsolver++", "use_karras_sigmas": False},
    ),
    (
        Scheduler.DPM_PP_2M_K,
        {"algorithm_type": "dpmsolver++", "use_karras_sigmas": True},
    ),
    (
        Scheduler.DPM_PP_2M_SDE_K,
        {"algorithm_type": "sde-dpmsolver++", "use_karras_sigmas": True},
    ),
    (
        Scheduler.DPM_2M_SDE_K,
        {"algorithm_type": "sde-dpmsolver", "use_karras_sigmas": True},
    ),
]

for scheduler_enum, expected_overrides in test_cases:
    scheduler_name = scheduler_enum.value
    print(f"\n{scheduler_name}:")
    print(f"  Expected overrides: {expected_overrides}")

print("\n" + "=" * 60)
print("Done! If you see checkmarks above, the scheduler database is correct.")
print("Now run airunner and check the logs for scheduler change messages.")
print("=" * 60)



# Copyright 2025 Alibaba Z-Image Team and The HuggingFace Team. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# This file is adapted from the diffusers library main branch for airunner.
# Once ZImagePipeline is available in a stable diffusers release, this
# local copy should be removed and the diffusers version used instead.

from dataclasses import dataclass
from typing import List, Union

import numpy as np
import PIL.Image

from diffusers.utils import BaseOutput


@dataclass
class ZImagePipelineOutput(BaseOutput):
    """
    Output class for Z-Image pipelines.

    Args:
        images (`List[PIL.Image.Image]` or `np.ndarray`)
            List of denoised PIL images of length `batch_size` or numpy array of shape
            `(batch_size, height, width, num_channels)`. PIL images or numpy array present
            the denoised images of the diffusion pipeline.
    """

    images: Union[List[PIL.Image.Image], np.ndarray]



"""
LLM-Guided Spider for intelligent web crawling.

This spider uses external LLM decision-making to determine which links
to follow during crawling. Instead of blindly following all links, it
yields page data to an external controller (LLM) and waits for decisions
on which links are worth following.
"""

import scrapy
from typing import Callable, Optional, Set, Dict, List

from airunner.components.tools.web_content_extractor import WebContentExtractor
from airunner.settings import AIRUNNER_LOG_LEVEL
from airunner.utils.application import get_logger

logger = get_logger(__name__, AIRUNNER_LOG_LEVEL)


class LLMGuidedSpider(scrapy.Spider):
    """
    Scrapy spider that uses LLM decision-making for intelligent crawling.

    This spider:
    1. Scrapes a web page
    2. Extracts clean content + links
    3. Yields page data to external decision callback
    4. Receives list of URLs to follow from callback
    5. Continues crawling selected URLs
    """

    name = "llm_guided"

    def __init__(
        self,
        start_url: str,
        decision_callback: Optional[Callable] = None,
        research_goal: str = "",
        max_pages: int = 20,
        max_depth: int = 3,
        *args,
        **kwargs,
    ):
        """
        Initialize the LLM-guided spider.

        Args:
            start_url: Initial URL to start crawling from
            decision_callback: Function that receives page data and returns
                             dict with 'follow_urls' list
            research_goal: Description of what we're researching (for context)
            max_pages: Maximum number of pages to crawl
            max_depth: Maximum crawl depth
        """
        super().__init__(*args, **kwargs)

        self.start_urls = [start_url]
        self.decision_callback = decision_callback
        self.research_goal = research_goal
        self.max_pages = max_pages
        self.max_depth = max_depth

        # State tracking
        self.visited: Set[str] = set()
        self.pages_scraped: int = 0
        self.collected_content: List[Dict] = []

        # Configure custom settings
        self.custom_settings = {
            "DEPTH_LIMIT": max_depth,
            "CLOSESPIDER_PAGECOUNT": max_pages,
        }

        logger.info(
            f"LLMGuidedSpider initialized: goal='{research_goal}', "
            f"max_pages={max_pages}, max_depth={max_depth}"
        )

    def parse(self, response):
        """
        Parse a web page: extract content, get LLM decision, follow links.

        This is the main parsing method called for each crawled page.
        """
        url = response.url
        self.visited.add(url)
        self.pages_scraped += 1

        logger.info(
            f"Parsing page {self.pages_scraped}/{self.max_pages}: {url}"
        )

        # Extract content and links
        page_data = self._extract_page_data(response)

        if not page_data:
            logger.warning(f"Failed to extract data from {url}")
            return

        # Store the collected content
        self.collected_content.append(page_data)

        # Get LLM decision on which links to follow
        if self.decision_callback and self.pages_scraped < self.max_pages:
            try:
                decision = self.decision_callback(
                    page_data=page_data,
                    research_goal=self.research_goal,
                    pages_scraped=self.pages_scraped,
                    max_pages=self.max_pages,
                )

                # Extract URLs to follow from decision
                urls_to_follow = decision.get("follow_urls", [])

                logger.info(
                    f"LLM decision: follow {len(urls_to_follow)} links, "
                    f"sufficient_info={decision.get('sufficient_info', False)}"
                )

                # Stop crawling if LLM says we have enough info
                if decision.get("sufficient_info", False):
                    logger.info(
                        "LLM indicated sufficient information gathered - stopping crawl"
                    )
                    return

                # Follow approved URLs
                for link_url in urls_to_follow[:3]:  # Limit to top 3
                    if link_url not in self.visited:
                        logger.info(f"Following LLM-selected link: {link_url}")
                        yield scrapy.Request(
                            link_url,
                            callback=self.parse,
                            errback=self.errback_httpbin,
                            dont_filter=False,  # Still respect Scrapy's duplicate filter
                        )

            except Exception as e:
                logger.error(f"Error in decision callback: {e}", exc_info=True)

        # Yield the scraped page data
        yield page_data

    def _extract_page_data(self, response) -> Optional[Dict]:
        """
        Extract content, links, and metadata from a response.

        Args:
            response: Scrapy response object

        Returns:
            Dictionary with content, links, metadata, or None if extraction fails
        """
        try:
            url = response.url
            html_content = response.text

            # Use WebContentExtractor to get content + links
            result = WebContentExtractor.extract_with_links(
                url, content=html_content
            )

            if not result:
                logger.warning(f"WebContentExtractor returned None for {url}")
                return None

            # Add URL and depth info
            result["url"] = url
            result["depth"] = response.meta.get("depth", 0)
            result["page_number"] = self.pages_scraped

            logger.info(
                f"Extracted {len(result['content'])} chars content, "
                f"{len(result['links'])} links from {url}"
            )

            return result

        except Exception as e:
            logger.error(
                f"Error extracting page data from {response.url}: {e}",
                exc_info=True,
            )
            return None

    def errback_httpbin(self, failure):
        """Handle request failures."""
        logger.error(f"Request failed: {repr(failure)}")

    def closed(self, reason):
        """Called when spider closes."""
        logger.info(
            f"Spider closed: {reason}. "
            f"Scraped {self.pages_scraped} pages, "
            f"collected {len(self.collected_content)} page data items"
        )



import sys


PLATFORM_WINDOWS = "windows"
PLATFORM_LINUX = "linux"
PLATFORM_BSD = "bsd"
PLATFORM_DARWIN = "darwin"
PLATFORM_UNKNOWN = "unknown"


def get_platform_name():
    if sys.platform.startswith("win"):
        return PLATFORM_WINDOWS
    elif sys.platform.startswith("darwin"):
        return PLATFORM_DARWIN
    elif sys.platform.startswith("linux"):
        return PLATFORM_LINUX
    elif sys.platform.startswith(
        ("dragonfly", "freebsd", "netbsd", "openbsd", "bsd")
    ):
        return PLATFORM_BSD
    else:
        return PLATFORM_UNKNOWN


__platform__ = get_platform_name()


def is_linux():
    return __platform__ == PLATFORM_LINUX


def is_bsd():
    return __platform__ == PLATFORM_BSD


def is_darwin():
    return __platform__ == PLATFORM_DARWIN


def is_windows():
    return __platform__ == PLATFORM_WINDOWS



from PySide6 import QtCore
from airunner.vendor.nodegraphqt.widgets.viewer import NodeViewer
from airunner.components.application.api.api import API


class DebouncedNodeViewer(NodeViewer):
    """
    Extended NodeViewer with debounced zoom and pan signals that emit to the API.
    """

    def __init__(self, parent=None, undo_stack=None):
        super(DebouncedNodeViewer, self).__init__(parent, undo_stack)

        # Initialize API reference
        self.api = API()

        # Create debounce timers for zoom and pan events
        self._zoom_timer = QtCore.QTimer()
        self._zoom_timer.setSingleShot(True)
        self._zoom_timer.timeout.connect(self._emit_zoom_changed)

        self._pan_timer = QtCore.QTimer()
        self._pan_timer.setSingleShot(True)
        self._pan_timer.timeout.connect(self._emit_pan_changed)

        # Timer delays in milliseconds
        self._debounce_delay = 250  # 250ms debouncing

        self.zooms = 0
        self.pan = (0, 0)

        # Ensure _zoom and _pan are initialized for property access
        self._zoom = 1.0
        self._pan = (0, 0)

    @property
    def zoom(self):
        return self._zoom

    @zoom.setter
    def zoom(self, value):
        self._zoom = value

    @property
    def pan(self):
        return self._pan

    @pan.setter
    def pan(self, value):
        self._pan = value

    def _set_viewer_zoom(self, value, sensitivity=None, pos=None):
        """Override the zoom method to add debouncing signal."""
        super()._set_viewer_zoom(value, sensitivity, pos)

        self._zoom = value

        # Start or restart the debounce timer
        self._zoom_timer.start(self._debounce_delay)

    def _set_viewer_pan(self, pos_x, pos_y):
        """Override the pan method to add debouncing signal."""
        super()._set_viewer_pan(pos_x, pos_y)

        self._pan = (pos_x, pos_y)

        # Start or restart the debounce timer
        self._pan_timer.start(self._debounce_delay)

    def _emit_zoom_changed(self):
        """Emit the zoom changed signal after debouncing."""
        zoom_level = self.get_zoom()
        self.api.nodegraph.zoom_changed(zoom_level)

    def _emit_pan_changed(self):
        """Emit the pan changed signal after debouncing."""
        center = self.scene_center()
        self.api.nodegraph.pan_changed(int(center[0]), int(center[1]))



"""Smoke tests for RAGMixin refactoring.

These tests verify that the refactored RAGMixin works correctly after
extracting into 6 mixins + retriever class.
"""

from unittest.mock import Mock


class TestRAGMixinRefactoring:
    """Smoke tests to verify RAGMixin refactoring."""

    def test_rag_mixin_imports_successfully(self):
        """Should be able to import RAGMixin after refactoring."""
        from airunner.components.llm.managers.agent.rag_mixin import RAGMixin

        assert RAGMixin is not None

    def test_rag_mixin_has_correct_mro(self):
        """Should have all 6 mixins in method resolution order."""
        from airunner.components.llm.managers.agent.rag_mixin import RAGMixin

        mro_names = [cls.__name__ for cls in RAGMixin.__mro__]

        assert "RAGPropertiesMixin" in mro_names
        assert "RAGDocumentMixin" in mro_names
        assert "RAGIndexManagementMixin" in mro_names
        assert "RAGIndexingMixin" in mro_names
        assert "RAGSearchMixin" in mro_names
        assert "RAGLifecycleMixin" in mro_names

    def test_rag_mixin_has_all_public_methods(self):
        """Should have all expected public methods from mixins."""
        from airunner.components.llm.managers.agent.rag_mixin import RAGMixin

        expected_methods = [
            # From RAGPropertiesMixin
            "text_splitter",
            "doc_indexes_dir",
            "registry_path",
            "index_registry",
            "embedding",
            "target_files",
            "rag_system_prompt",
            "storage_persist_dir",
            # From RAGDocumentMixin
            # (all private methods)
            # From RAGSearchMixin
            "search",
            "get_retriever_for_query",
            "retriever",
            # From RAGIndexingMixin
            "document_reader",
            "documents",
            "index_all_documents",
            # From RAGLifecycleMixin
            "reload_rag",
            "clear_rag_documents",
            "unload_rag",
            "load_html_into_rag",
        ]

        for method in expected_methods:
            assert hasattr(RAGMixin, method), f"Missing method: {method}"

    def test_multi_index_retriever_imports_successfully(self):
        """Should be able to import MultiIndexRetriever."""
        from airunner.components.llm.managers.agent.retriever import (
            MultiIndexRetriever,
        )

        assert MultiIndexRetriever is not None

    def test_all_mixins_import_successfully(self):
        """Should be able to import all individual mixins."""
        from airunner.components.llm.managers.agent.mixins import (
            RAGPropertiesMixin,
            RAGDocumentMixin,
            RAGIndexManagementMixin,
            RAGIndexingMixin,
            RAGSearchMixin,
            RAGLifecycleMixin,
        )

        assert RAGPropertiesMixin is not None
        assert RAGDocumentMixin is not None
        assert RAGIndexManagementMixin is not None
        assert RAGIndexingMixin is not None
        assert RAGSearchMixin is not None
        assert RAGLifecycleMixin is not None

    def test_rag_mixin_can_be_instantiated_with_mocks(self):
        """Should be able to create instance with mocked dependencies."""
        from airunner.components.llm.managers.agent.rag_mixin import RAGMixin

        class TestRAG(RAGMixin):
            def __init__(self):
                self.logger = Mock()
                self.path_settings = Mock()
                self.path_settings.base_path = "/tmp/test"
                self.system_prompt = "Test"
                self.botname = "TestBot"
                self._skip_agent_load = True  # Skip embedding setup
                super().__init__()

        # Should not raise
        instance = TestRAG()
        assert instance is not None

    def test_target_files_property_works(self):
        """Should be able to get/set target_files property."""
        from airunner.components.llm.managers.agent.rag_mixin import RAGMixin

        class TestRAG(RAGMixin):
            def __init__(self):
                self.logger = Mock()
                self.path_settings = Mock()
                self.path_settings.base_path = "/tmp/test"
                self.system_prompt = "Test"
                self.botname = "TestBot"
                self._skip_agent_load = True
                super().__init__()

        instance = TestRAG()

        # Should be able to set
        instance.target_files = ["file1.txt", "file2.txt"]
        assert instance.target_files == ["file1.txt", "file2.txt"]

        # Should be able to clear
        instance.target_files = None
        assert instance.target_files is None

    def test_reload_rag_clears_state(self):
        """Should clear all caches when reload_rag is called."""
        from airunner.components.llm.managers.agent.rag_mixin import RAGMixin

        class TestRAG(RAGMixin):
            def __init__(self):
                self.logger = Mock()
                self.path_settings = Mock()
                self.path_settings.base_path = "/tmp/test"
                self.system_prompt = "Test"
                self.botname = "TestBot"
                self._skip_agent_load = True
                super().__init__()

        instance = TestRAG()

        # Set some state
        instance._index = Mock()
        instance._retriever = Mock()
        instance._doc_metadata_cache["test"] = "data"

        # Reload should clear
        instance.reload_rag()

        assert instance._index is None
        assert instance._retriever is None
        assert len(instance._doc_metadata_cache) == 0



from airunner.components.downloader.gui.windows.setup_wizard.base_wizard import BaseWizard
from airunner.components.downloader.gui.windows.setup_wizard.installation_settings.templates.install_success_page_ui import (
    Ui_install_success_page,
)


class InstallSuccessPage(BaseWizard):
    class_name_ = Ui_install_success_page



from PySide6.QtCore import (
    QPointF,
    Qt,
)
from PySide6.QtWidgets import (
    QGraphicsTextItem,
    QMenu,
)
from airunner.enums import CanvasToolName
from airunner.components.art.utils.canvas_position_manager import (
    CanvasPositionManager,
    ViewState,
)


class DraggableTextItem(QGraphicsTextItem):
    """A text item that supports dragging in MOVE tool and reports position changes back to the view."""

    def __init__(self, view: "CustomGraphicsView"):
        super().__init__()
        self._view = view
        self.initial_mouse_scene_pos = None
        self.initial_item_abs_pos = None
        self.mouse_press_pos = None
        self._current_snapped_pos = (0, 0)

    @property
    def current_tool(self):
        try:
            return self._view.current_tool
        except Exception:
            return None

    def mousePressEvent(self, event):
        # Only initiate custom drag when in MOVE tool
        if self.current_tool not in [CanvasToolName.MOVE]:
            return QGraphicsTextItem.mousePressEvent(self, event)

        if event.button() == Qt.MouseButton.LeftButton:
            self.initial_mouse_scene_pos = event.scenePos()

            # Use the item's current scene position to compute initial absolute
            # coordinates for dragging. Previously this used persisted layer
            # settings which could be stale, causing the item to jump when the
            # user started dragging. Using scenePos() (plus canvas offset)
            # reflects the true displayed position and prevents the jump.
            try:
                item_scene_pos = self.scenePos()
                canvas_offset = self._view.canvas_offset
                abs_x = item_scene_pos.x() + canvas_offset.x()
                abs_y = item_scene_pos.y() + canvas_offset.y()
            except Exception:
                abs_x = int(self.x() + self._view.canvas_offset_x)
                abs_y = int(self.y() + self._view.canvas_offset_y)

            self.initial_item_abs_pos = QPointF(abs_x, abs_y)
            self.mouse_press_pos = event.pos()
            event.accept()
        else:
            # Show a simple context menu on right-click to delete the text
            if event.button() == Qt.MouseButton.RightButton:
                try:
                    menu = QMenu()
                    delete_action = menu.addAction(self.tr("Delete"))
                    chosen = menu.exec_(event.screenPos())
                    if chosen == delete_action:
                        view = getattr(self, "_view", None)
                        if view is not None:
                            view._remove_text_item(self)
                        return
                except Exception:
                    pass
                return
            return QGraphicsTextItem.mousePressEvent(self, event)

    def mouseMoveEvent(self, event):
        if self.current_tool not in [CanvasToolName.MOVE]:
            return QGraphicsTextItem.mouseMoveEvent(self, event)

        if self.initial_mouse_scene_pos is not None:
            delta = event.scenePos() - self.initial_mouse_scene_pos
            proposed_abs_pos = self.initial_item_abs_pos + delta

            # Get view state
            view_state = ViewState(
                canvas_offset=self._view.canvas_offset,
                grid_compensation=getattr(
                    self._view, "grid_compensation_offset", QPointF(0, 0)
                ),
            )

            manager = CanvasPositionManager()

            # Snap to grid if enabled
            if self._view.grid_settings.snap_to_grid:
                snapped_abs_pos = manager.snap_to_grid(
                    proposed_abs_pos, self._view.grid_settings.cell_size
                )
            else:
                snapped_abs_pos = proposed_abs_pos

            # Convert to display position
            display_pos = manager.absolute_to_display(
                snapped_abs_pos, view_state
            )
            self.setPos(display_pos)
            self._current_snapped_pos = (
                int(snapped_abs_pos.x()),
                int(snapped_abs_pos.y()),
            )
            event.accept()
        else:
            return QGraphicsTextItem.mouseMoveEvent(self, event)

    def mouseReleaseEvent(self, event):
        if self.current_tool not in [CanvasToolName.MOVE]:
            return QGraphicsTextItem.mouseReleaseEvent(self, event)

        if self.initial_mouse_scene_pos is not None:
            has_moved = False
            if self.mouse_press_pos:
                has_moved = (
                    self.mouse_press_pos.x() != event.pos().x()
                    or self.mouse_press_pos.y() != event.pos().y()
                )

            self.initial_mouse_scene_pos = None
            self.initial_item_abs_pos = None
            self.mouse_press_pos = None

            if has_moved:
                # Persist via view's save routine (which groups by layer)
                try:
                    self._view._save_text_items_to_db()
                except Exception:
                    # fallback to calling public method
                    try:
                        self._view._save_text_items_to_db()
                    except Exception:
                        pass
            event.accept()
        else:
            return QGraphicsTextItem.mouseReleaseEvent(self, event)



"""Model Selector Widget with VSCode-style search for provider and model selection.

Provides an improved UI for selecting LLM providers and models with:
- Searchable provider dropdown
- Searchable model dropdown with details
- Model information display (VRAM, context length, capabilities)
- Quick access to Manage Models dialog
"""

from typing import Optional
from PySide6.QtWidgets import (
    QWidget,
    QVBoxLayout,
    QHBoxLayout,
    QLabel,
    QPushButton,
    QGroupBox,
    QFormLayout,
)
from PySide6.QtCore import Signal, Slot

from airunner.components.application.gui.widgets.searchable_combo_box import (
    SearchableComboBox,
)
from airunner.components.llm.config.provider_config import LLMProviderConfig
from airunner.enums import ModelService


class ModelSelectorWidget(QWidget):
    """Widget for selecting LLM provider and model with enhanced UI.

    Features:
    - Searchable provider dropdown
    - Searchable model dropdown with filtering
    - Model information display (VRAM, context, capabilities)
    - Manage Models button for advanced operations
    """

    # Signals
    provider_changed = Signal(str)  # Emits provider name
    model_changed = Signal(str, str)  # Emits (provider, model_id)

    def __init__(self, parent: Optional[QWidget] = None):
        """Initialize model selector widget.

        Args:
            parent: Parent widget
        """
        super().__init__(parent)
        self._current_provider: Optional[str] = None
        self._current_model_id: Optional[str] = None
        self._setup_ui()
        self._populate_providers()

    def _setup_ui(self):
        """Set up the widget UI."""
        layout = QVBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)

        # Provider selection
        provider_group = QGroupBox("Provider")
        provider_layout = QVBoxLayout(provider_group)

        self.provider_combo = SearchableComboBox()
        self.provider_combo.setPlaceholderText("Search providers...")
        self.provider_combo.itemSelected.connect(self._on_provider_selected)
        provider_layout.addWidget(self.provider_combo)

        layout.addWidget(provider_group)

        # Model selection
        model_group = QGroupBox("Model")
        model_layout = QVBoxLayout(model_group)

        self.model_combo = SearchableComboBox()
        self.model_combo.setPlaceholderText("Search models...")
        self.model_combo.itemSelected.connect(self._on_model_selected)
        model_layout.addWidget(self.model_combo)

        # Model info display
        info_layout = QFormLayout()

        self.vram_label = QLabel("-")
        self.vram_label.setStyleSheet("color: #666;")
        info_layout.addRow("VRAM (4-bit):", self.vram_label)

        self.context_label = QLabel("-")
        self.context_label.setStyleSheet("color: #666;")
        info_layout.addRow("Context:", self.context_label)

        self.tools_label = QLabel("-")
        self.tools_label.setStyleSheet("color: #666;")
        info_layout.addRow("Tool Calling:", self.tools_label)

        model_layout.addLayout(info_layout)

        # Manage Models button
        button_layout = QHBoxLayout()
        button_layout.addStretch()

        self.manage_button = QPushButton("Manage Models...")
        self.manage_button.clicked.connect(self._on_manage_clicked)
        button_layout.addWidget(self.manage_button)

        model_layout.addLayout(button_layout)

        layout.addWidget(model_group)
        layout.addStretch()

    def _populate_providers(self):
        """Populate provider dropdown with available providers."""
        self.provider_combo.clear()

        providers = [
            (
                "Local Models",
                ModelService.LOCAL.value,
                "Run models locally on your hardware",
            ),
            (
                "OpenRouter",
                ModelService.OPENROUTER.value,
                "Access multiple providers through OpenRouter",
            ),
            (
                "Ollama",
                ModelService.OLLAMA.value,
                "Use Ollama for local model serving",
            ),
        ]

        self.provider_combo.addItemsWithData(providers)

    def _populate_models(self, provider: str):
        """Populate model dropdown based on selected provider.

        Args:
            provider: Provider name
        """
        self.model_combo.clear()

        models = LLMProviderConfig.get_models_for_provider(provider)

        for model_id in models:
            display_name = LLMProviderConfig.get_model_display_name(
                provider, model_id
            )

            # Get model info for tooltip
            model_info = LLMProviderConfig.get_model_info(provider, model_id)
            tooltip = None
            if model_info and model_info.get("description"):
                tooltip = model_info["description"]

            if tooltip:
                self.model_combo.addItemWithData(
                    display_name, model_id, tooltip
                )
            else:
                self.model_combo.addItemWithData(display_name, model_id)

    def _update_model_info(self, provider: str, model_id: str):
        """Update model information display.

        Args:
            provider: Provider name
            model_id: Model identifier
        """
        if provider != ModelService.LOCAL.value:
            # For remote providers, show generic info
            self.vram_label.setText("N/A (remote)")
            self.context_label.setText("Varies by model")
            self.tools_label.setText("Check provider docs")
            return

        model_info = LLMProviderConfig.get_model_info(provider, model_id)
        if not model_info or model_id == "custom":
            self.vram_label.setText("-")
            self.context_label.setText("-")
            self.tools_label.setText("-")
            return

        # VRAM info
        vram_4bit = model_info.get("vram_4bit_gb", 0)
        if vram_4bit > 0:
            self.vram_label.setText(f"~{vram_4bit} GB")
        else:
            self.vram_label.setText("Unknown")

        # Context length
        context = model_info.get("context_length", 0)
        if context > 0:
            if context >= 1000000:
                self.context_label.setText(f"{context // 1000000}M tokens")
            elif context >= 1000:
                self.context_label.setText(f"{context // 1000}K tokens")
            else:
                self.context_label.setText(f"{context} tokens")
        else:
            self.context_label.setText("Unknown")

        # Tool calling support
        has_tools = model_info.get("function_calling", False)
        tool_mode = model_info.get("tool_calling_mode", "none")
        if has_tools:
            self.tools_label.setText(f"Yes ({tool_mode})")
        else:
            self.tools_label.setText("No (chat only)")

    @Slot(object)
    def _on_provider_selected(self, provider_data: str):
        """Handle provider selection.

        Args:
            provider_data: Selected provider name
        """
        self._current_provider = provider_data
        self._populate_models(provider_data)
        self.provider_changed.emit(provider_data)

        # Clear model info
        self._update_model_info("", "")

    @Slot(object)
    def _on_model_selected(self, model_data: str):
        """Handle model selection.

        Args:
            model_data: Selected model ID
        """
        if not self._current_provider:
            return

        self._current_model_id = model_data
        self._update_model_info(self._current_provider, model_data)
        self.model_changed.emit(self._current_provider, model_data)

    @Slot()
    def _on_manage_clicked(self):
        """Handle Manage Models button click."""
        from airunner.components.settings.gui.widgets.model_manager_dialog import (
            ManageModelsDialog,
        )

        dialog = ManageModelsDialog(self)

        # Connect signals if needed (download/delete)
        # These would typically be connected to actual download/delete handlers

        dialog.exec()

    def set_provider(self, provider: str):
        """Set current provider.

        Args:
            provider: Provider name
        """
        self.provider_combo.setCurrentData(provider)
        self._current_provider = provider
        self._populate_models(provider)

    def set_model(self, model_id: str):
        """Set current model.

        Args:
            model_id: Model identifier
        """
        self.model_combo.setCurrentData(model_id)
        self._current_model_id = model_id

        if self._current_provider:
            self._update_model_info(self._current_provider, model_id)

    def get_provider(self) -> Optional[str]:
        """Get currently selected provider.

        Returns:
            Provider name or None
        """
        return self._current_provider

    def get_model_id(self) -> Optional[str]:
        """Get currently selected model ID.

        Returns:
            Model ID or None
        """
        return self._current_model_id
